#!/usr/bin/env python3

# ResourceAssigner.py: ResourceAssigner listens on the lofar ?? bus and calls onTaskSpecified
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: raservice.py 1580 2015-09-30 14:18:57Z loose $

"""
TaskSpecifiedListener listens to a bus on which specified tasks get published. It will then try
to assign resources to these tasks.
"""

import logging

from lofar.common import dbcredentials
from lofar.messaging import DEFAULT_BROKER, DEFAULT_BUSNAME
from lofar.messaging.rpc import RPCService, ServiceMessageHandler
from lofar.sas.resourceassignment.rataskspecified.RABusListener import RATaskSpecifiedEventMessageHandler, RATaskSpecifiedBusListener
from lofar.sas.resourceassignment.resourceassigner.resource_assigner import ResourceAssigner
from lofar.sas.resourceassignment.resourceassigner.schedulechecker import ScheduleChecker

logger = logging.getLogger(__name__)


class SpecifiedTaskEventMessageHandler(RATaskSpecifiedEventMessageHandler):
    def __init__(self, assigner=None):
        """
        SpecifiedTaskEventMessageHandler listens on the lofar ?? bus and calls onTaskSpecified
        """
        super().__init__()

        self.assigner = assigner
        if not self.assigner:
            self.assigner = ResourceAssigner()

    def onTaskSpecified(self, otdb_id, specification_tree):
        logger.info('onTaskSpecified: otdb_id=%s status=%s', otdb_id, specification_tree.get('status', '').lower())

        try:
            self.assigner.do_assignment(specification_tree)
        except Exception as e:
            logger.error(str(e))

class ResourceAssignerServiceMessageHandler(ServiceMessageHandler):
    def __init__(self, assigner=None):
        super().__init__()
        self.assigner = assigner

    def do_assignment(self, specification_tree):
        return self.assigner.do_assignment(specification_tree)

class RAService:
    def __init__(self, radbcreds: dbcredentials.DBCredentials=None, exchange: str = DEFAULT_BUSNAME, broker: str = DEFAULT_BROKER):
        if radbcreds is None:
            radbcreds = dbcredentials.DBCredentials().get("RADB")
            logger.info("Read default RADB dbcreds from disk: %s" % radbcreds.stringWithHiddenPassword())

        self.assigner = ResourceAssigner(exchange=exchange, broker=broker, radb_dbcreds=radbcreds)

        # create a buslistener for the Event-Driven paradigm, doing resourceassignment upon receiving a RATaskSpecified-EventMessage
        self.rpcservice = RPCService(service_name="RAService",
                                     handler_type=ResourceAssignerServiceMessageHandler,
                                     handler_kwargs={"assigner": self.assigner},
                                     exchange=exchange,
                                     broker=broker)

        # create a service for the blocking RPC call, doing resourceassignment upon receiving a RequestMessage
        self.rataskspecifiedbuslistener = RATaskSpecifiedBusListener(handler_type=SpecifiedTaskEventMessageHandler,
                                                                     handler_kwargs={"assigner": self.assigner},
                                                                     exchange=exchange,
                                                                     broker=broker)

    def start_listening(self):
        self.assigner.open()
        self.rpcservice.start_listening()
        self.rataskspecifiedbuslistener.start_listening()

    def stop_listening(self):
        self.rpcservice.stop_listening()
        self.rataskspecifiedbuslistener.stop_listening()
        self.assigner.close()

    def __enter__(self):
        try:
            self.start_listening()
            return self
        except Exception as e:
            # __exit__ (and hence stop_listening) is not called when an exception is raised in __enter__
            # so, do our own cleanup (log, stop_listening and re-raise).
            logger.exception("%s error: %s",  self, e)
            self.stop_listening()
            raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_listening()



def main():
    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    from optparse import OptionParser
    from lofar.common.util import waitForInterrupt

    # Check the invocation arguments
    parser = OptionParser("%prog [options]",
                          description='runs the resourceassigner service')
    parser.add_option('-b', '--broker', dest='broker', type='string',
                      default=DEFAULT_BROKER,
                      help='Address of the broker, default: localhost')
    parser.add_option('-e', "--exchange", dest="exchange", type="string",
                      default=DEFAULT_BUSNAME,
                      help="Name of the bus on which communication occurs [default: %default]")
    parser.add_option('-V', '--verbose', dest='verbose', action='store_true', help='verbose logging')

    parser.add_option_group(dbcredentials.options_group(parser))
    parser.set_defaults(dbcredentials="RADB")

    (options, args) = parser.parse_args()

    radbcreds = dbcredentials.parse_options(options)

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.DEBUG if options.verbose else logging.INFO)

    with RAService(radbcreds=radbcreds, exchange=options.exchange, broker=options.broker):
        with ScheduleChecker(exchange=options.exchange, broker=options.broker):
            waitForInterrupt()

if __name__ == '__main__':
    main()
