#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import uuid
from unittest import mock

from astropy.coordinates import Angle

import logging
logger = logging.getLogger(__name__)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

TEST_UUID = uuid.uuid1()

from datetime import datetime, timedelta
from lofar.common.json_utils import get_default_json_object_for_schema, add_defaults_to_json_object_for_schema
from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor

tmp_exchange = TemporaryExchange("t_dynamic_scheduling_%s" % (TEST_UUID,))
tmp_exchange.open()

# override DEFAULT_BUSNAME
import lofar
lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address

from lofar.sas.tmss.test.test_utils import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(exchange=tmp_exchange.address,
                                    populate_schemas=True, populate_test_data=False,
                                    start_postgres_listener=True, start_subtask_scheduler=False,
                                    start_ra_test_environment=True, enable_viewflow=False,
                                    start_dynamic_scheduler=False)  # do not start the dynamic scheduler in the testenv, because it is the object-under-test.
tmss_test_env.start()
from django.test import TestCase

def tearDownModule():
    tmss_test_env.stop()
    tmp_exchange.close()

from lofar.sas.tmss.test.tmss_test_data_django_models import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprints_and_subtasks_from_scheduling_unit_draft
from lofar.sas.tmss.tmss.tmssapp.subtasks import unschedule_subtask
from lofar.common.postgres import PostgresDatabaseConnection

# the module under test
import lofar.sas.tmss.services.scheduling.constraints.template_constraints_v1 as tc1
from lofar.sas.tmss.services.scheduling.dynamic_scheduling import *


class TestDynamicScheduling(TestCase):  # Note: we use django.test.TestCase instead of unittest.TestCase to avoid manual cleanup of objects created by other tests
    '''
    Tests for the Dynamic Scheduling
    '''
    @classmethod
    def setUpClass(cls) -> None:
        super(TestDynamicScheduling, cls).setUpClass()
        # make some re-usable projects with high/low priority
        cls.project_low = models.Project.objects.create(**Project_test_data("dynamic scheduling test project %s"% (uuid.uuid4(),), priority_rank=1))
        cls.project_medium = models.Project.objects.create(**Project_test_data("dynamic scheduling test project %s"% (uuid.uuid4(),), priority_rank=2))
        cls.project_high = models.Project.objects.create(**Project_test_data("dynamic scheduling test project %s"% (uuid.uuid4(),), priority_rank=3))
        cls.scheduling_set_low = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project_low))
        cls.scheduling_set_medium = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project_medium))
        cls.scheduling_set_high = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project_high))

    def setUp(self) -> None:
        # wipe all radb entries (via cascading deletes) in between tests, so the tests don't influence each other
        with PostgresDatabaseConnection(tmss_test_env.ra_test_environment.radb_test_instance.dbcreds) as radb:
            radb.executeQuery('DELETE FROM resource_allocation.specification;')
            radb.executeQuery('TRUNCATE resource_allocation.resource_usage;')
            radb.commit()

        # wipe all scheduling_unit_drafts in between tests, so the tests don't influence each other
        for scheduling_set in [self.scheduling_set_low, self.scheduling_set_medium, self.scheduling_set_high]:
            for scheduling_unit_draft in scheduling_set.scheduling_unit_drafts.all():
                for scheduling_unit_blueprint in scheduling_unit_draft.scheduling_unit_blueprints.all():
                    for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                        for subtask in task_blueprint.subtasks.all():
                            try:
                                if subtask.state.value == models.SubtaskState.Choices.SCHEDULED.value:
                                    unschedule_subtask(subtask)
                            except Exception as e:
                                logger.exception(e)
                            for output in subtask.outputs.all():
                                for dataproduct in output.dataproducts.all():
                                    dataproduct.delete()
                                for consumer in output.consumers.all():
                                    consumer.delete()
                                output.delete()
                            for input in subtask.inputs.all():
                                input.delete()
                            subtask.delete()
                        task_blueprint.draft.delete()
                        task_blueprint.delete()
                    scheduling_unit_blueprint.delete()
                scheduling_unit_draft.delete()

    @staticmethod
    def create_simple_observation_scheduling_unit(name:str=None, scheduling_set=None,
                                                  obs_duration:int=60,
                                                  constraints=None):
        constraints_template = models.SchedulingConstraintsTemplate.objects.get(name="constraints")
        constraints = add_defaults_to_json_object_for_schema(constraints or {}, constraints_template.schema)

        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="Simple Observation")
        scheduling_unit_spec = add_defaults_to_json_object_for_schema(strategy_template.template,
                                                                      strategy_template.scheduling_unit_template.schema)
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['duration'] = obs_duration
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['station_groups'][0]['stations'] = ['CS001']

        # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
        return models.SchedulingUnitDraft.objects.create(name=name,
                                                         scheduling_set=scheduling_set,
                                                         requirements_template=strategy_template.scheduling_unit_template,
                                                         requirements_doc=scheduling_unit_spec,
                                                         observation_strategy_template=strategy_template,
                                                         scheduling_constraints_doc=constraints,
                                                         scheduling_constraints_template=constraints_template)

    @unittest.skip("FIX TEST, skipping it for now, see TODO comment in assign_start_stop_times_to_schedulable_scheduling_units")
    def test_three_simple_observations_no_constraints_different_project_priority(self):
        scheduling_unit_draft_low = self.create_simple_observation_scheduling_unit("scheduling unit low", scheduling_set=self.scheduling_set_low)
        scheduling_unit_blueprint_low = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_low)

        scheduling_unit_draft_medium = self.create_simple_observation_scheduling_unit("scheduling unit medium", scheduling_set=self.scheduling_set_medium)
        scheduling_unit_blueprint_medium = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_medium)

        scheduling_unit_draft_high = self.create_simple_observation_scheduling_unit("scheduling unit high", scheduling_set=self.scheduling_set_high)
        scheduling_unit_blueprint_high = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_high)

        # call the method-under-test.
        scheduled_scheduling_unit = do_dynamic_schedule()

        # we expect the scheduling_unit with the highest project rank to be scheduled first
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduling_unit_blueprint_high.id, scheduled_scheduling_unit.id)

        # check the results
        # we expect the sub_high to be scheduled
        scheduling_unit_blueprint_low.refresh_from_db()
        scheduling_unit_blueprint_medium.refresh_from_db()
        scheduling_unit_blueprint_high.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint_low.status, 'schedulable')
        self.assertEqual(scheduling_unit_blueprint_medium.status, 'schedulable')
        self.assertEqual(scheduling_unit_blueprint_high.status, 'scheduled')

        # check the scheduled subtask
        upcoming_scheduled_subtasks = models.Subtask.objects.filter(state__value='scheduled',
                                                                    task_blueprint__scheduling_unit_blueprint__in=(scheduling_unit_blueprint_low,
                                                                                                                   scheduling_unit_blueprint_medium,
                                                                                                                   scheduling_unit_blueprint_high)).all()
        self.assertEqual(1, upcoming_scheduled_subtasks.count())
        self.assertEqual(scheduling_unit_blueprint_high.id, upcoming_scheduled_subtasks[0].task_blueprint.scheduling_unit_blueprint.id)

        # check scheduling_unit_blueprint_low starts after the scheduled scheduling_unit_blueprint_high
        self.assertGreater(scheduling_unit_blueprint_low.start_time, scheduling_unit_blueprint_medium.start_time)
        self.assertGreater(scheduling_unit_blueprint_medium.start_time, scheduling_unit_blueprint_high.start_time)

        # ensure DEFAULT_INTER_OBSERVATION_GAP between them
        self.assertGreaterEqual(scheduling_unit_blueprint_medium.start_time - scheduling_unit_blueprint_high.stop_time, DEFAULT_INTER_OBSERVATION_GAP)
        self.assertGreaterEqual(scheduling_unit_blueprint_low.start_time - scheduling_unit_blueprint_medium.stop_time, DEFAULT_INTER_OBSERVATION_GAP)

    def test_time_bound_unit_wins_even_at_lower_priority(self):
        # create two schedule units, one with high one with low prio.
        # first create them without any further constraints, and check if high prio wins.
        scheduling_unit_draft_low = self.create_simple_observation_scheduling_unit("scheduling unit low", scheduling_set=self.scheduling_set_low)
        scheduling_unit_blueprint_low = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_low)

        scheduling_unit_draft_high = self.create_simple_observation_scheduling_unit("scheduling unit high", scheduling_set=self.scheduling_set_high)
        scheduling_unit_blueprint_high = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_high)

        now = datetime.utcnow()
        tomorrow = now+timedelta(days=1)

        # call the method-under-test.
        best_scored_scheduling_unit = find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high], now, tomorrow)

        # we expect the scheduling_unit with the highest project rank to be scheduled first
        self.assertEqual(scheduling_unit_blueprint_high.id, best_scored_scheduling_unit.scheduling_unit.id)

        #now update the low prio unit with a time constraint, "forcing" it to be run in a very thight upcoming time window.
        scheduling_unit_draft_low.scheduling_constraints_doc['time'] = { 'before': (now+scheduling_unit_draft_low.duration).isoformat()+'Z' }
        scheduling_unit_draft_low.save()
        scheduling_unit_blueprint_low.refresh_from_db()

        # call the method-under-test.
        best_scored_scheduling_unit = find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high], now, tomorrow)

        # now we expect the scheduling_unit with the lowest project rank to be scheduled first because it can only run within this limited timewindow
        self.assertEqual(scheduling_unit_draft_low.id, best_scored_scheduling_unit.scheduling_unit.id)


        #  update the low prio unit. enlarge the time window constraint a bit, so both low and high prio units can fit
        # this should result that the high prio goes first, and the low prio (which now fits as well) goes second
        scheduling_unit_draft_low.scheduling_constraints_doc['time'] = \
            { 'before': (now+scheduling_unit_draft_low.duration+scheduling_unit_draft_high.duration).isoformat()+'Z' }
        scheduling_unit_draft_low.save()
        scheduling_unit_blueprint_low.refresh_from_db()

        # call the method-under-test.
        best_scored_scheduling_unit = find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high], now, tomorrow)

        # now we expect the scheduling_unit with the lowest project rank to be scheduled first because it can only
        # run within this limited timewindow
        self.assertEqual(scheduling_unit_blueprint_low.id, best_scored_scheduling_unit.scheduling_unit.id)

        # call the method-under-test again but search after first unit (should return low prio unit)
        stop_time_of_first =  best_scored_scheduling_unit.start_time + best_scored_scheduling_unit.scheduling_unit.duration
        best_scored_scheduling_unit = find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high], stop_time_of_first, tomorrow)
        self.assertEqual(scheduling_unit_blueprint_low.id, best_scored_scheduling_unit.scheduling_unit.id)

    def test_manual_constraint_is_preventing_scheduling_unit_from_being_scheduled_dynamically(self):
        scheduling_unit_draft_manual = self.create_simple_observation_scheduling_unit("scheduling unit manual low", scheduling_set=self.scheduling_set_low,
                                                                                      constraints={'scheduler': 'manual'})
        scheduling_unit_blueprint_manual = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_manual)
        self.assertEqual(scheduling_unit_blueprint_manual.status, "schedulable")

        # call the method-under-test.
        scheduled_scheduling_unit = do_dynamic_schedule()

        # we expect no scheduling_unit to be scheduled, because the only one is set to 'manual' constraint
        self.assertIsNone(scheduled_scheduling_unit)

        # check the results
        scheduling_unit_blueprint_manual.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint_manual.status, 'schedulable')

    @unittest.skip("FIX TEST, skipping it for now,...something with manual scheduler ?")
    def test_manually_scheduled_blocking_dynamically_scheduled(self):
        scheduling_unit_draft_manual = self.create_simple_observation_scheduling_unit("scheduling unit manual low", scheduling_set=self.scheduling_set_low,
                                                                                      constraints={'scheduler': 'manual'})
        scheduling_unit_blueprint_manual = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_manual)
        self.assertEqual(scheduling_unit_blueprint_manual.status, "schedulable")

        schedule_independent_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint_manual, datetime.utcnow())
        self.assertEqual(scheduling_unit_blueprint_manual.status, "scheduled")

        scheduling_unit_draft_high = self.create_simple_observation_scheduling_unit("scheduling unit online high", scheduling_set=self.scheduling_set_high)
        scheduling_unit_blueprint_high = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_high)

        # call the method-under-test.
        scheduled_scheduling_unit = do_dynamic_schedule()

        # we expect the no scheduling_unit to be scheduled, because the manual is in the way -> Fix it
        self.assertIsNone(scheduled_scheduling_unit)

        # check the results
        # we expect the sub_high to be scheduled
        scheduling_unit_blueprint_high.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint_high.status, 'schedulable')

        # check scheduling_unit_blueprint_low starts after the scheduled scheduling_unit_blueprint_high
        self.assertGreater(scheduling_unit_blueprint_high.start_time, scheduling_unit_blueprint_manual.start_time)

        # ensure DEFAULT_INTER_OBSERVATION_GAP between them
        self.assertGreaterEqual(scheduling_unit_blueprint_high.start_time - scheduling_unit_blueprint_manual.stop_time, DEFAULT_INTER_OBSERVATION_GAP)


class TestDailyConstraints(TestCase):
    '''
    Tests for the constraint checkers used in dynamic scheduling
    '''

    def setUp(self) -> None:
        # scheduling unit
        self.obs_duration = 120 * 60
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_unit_draft = TestDynamicScheduling.create_simple_observation_scheduling_unit("scheduling unit for ...%s" % self._testMethodName[30:],
                                                                                                scheduling_set=scheduling_set,
                                                                                                obs_duration=self.obs_duration)
        self.scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # mock out conversions for speedup and assertable timestamps
        # earliest_start_time requests timestamp and timestamp+1day
        self.sunrise_data = {
            'CS001': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}, {"start": datetime(2020, 1, 2, 7, 30, 0), "end": datetime(2020, 1, 2, 9, 30, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}, {"start": datetime(2020, 1, 2, 9, 30, 0), "end": datetime(2020, 1, 2, 15, 30, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)},{"start": datetime(2020, 1, 2, 15, 30, 0), "end": datetime(2020, 1, 2, 17, 30, 0)}],
                      "night": [{"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}, {"start": datetime(2020, 1, 2, 17, 30, 0), "end": datetime(2020, 1, 3, 7, 30, 0)}]},
            'DE601': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 45, 0), "end": datetime(2020, 1, 1, 9, 45, 0)}, {"start": datetime(2020, 1, 2, 7, 45, 0), "end": datetime(2020, 1, 2, 9, 45, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 45, 0), "end": datetime(2020, 1, 1, 15, 45, 0)}, {"start": datetime(2020, 1, 2, 9, 45, 0), "end": datetime(2020, 1, 2, 15, 45, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 45, 0), "end": datetime(2020, 1, 1, 17, 45, 0)}, {"start": datetime(2020, 1, 2, 15, 45, 0), "end": datetime(2020, 1, 2, 17, 45, 0)}],
                      "night": [{"start": datetime(2020, 1, 1, 17, 45, 0), "end": datetime(2020, 1, 2, 7, 45, 0)}, {"start": datetime(2020, 1, 2, 17, 45, 0), "end": datetime(2020, 1, 3, 7, 45, 0)}]}}

        # variant for timestamp before sunrise, which returns the previous night
        self.sunrise_data_early_night = {
            'CS001': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}, {"start": datetime(2020, 1, 2, 7, 30, 0), "end": datetime(2020, 1, 2, 9, 30, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}, {"start": datetime(2020, 1, 2, 9, 30, 0), "end": datetime(2020, 1, 2, 15, 30, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)},{"start": datetime(2020, 1, 2, 15, 30, 0), "end": datetime(2020, 1, 2, 17, 30, 0)}],
                      "night": [{"start": datetime(2019, 12, 31, 17, 30, 0), "end": datetime(2020, 1, 1, 7, 30, 0)}, {"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}]},
            'DE601': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 45, 0), "end": datetime(2020, 1, 1, 9, 45, 0)}, {"start": datetime(2020, 1, 2, 7, 45, 0), "end": datetime(2020, 1, 2, 9, 45, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 45, 0), "end": datetime(2020, 1, 1, 15, 45, 0)}, {"start": datetime(2020, 1, 2, 9, 45, 0), "end": datetime(2020, 1, 2, 15, 45, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 45, 0), "end": datetime(2020, 1, 1, 17, 45, 0)},{"start": datetime(2020, 1, 2, 15, 45, 0), "end": datetime(2020, 1, 2, 17, 45, 0)}],
                      "night": [{"start": datetime(2019, 12, 31, 17, 45, 0), "end": datetime(2020, 1, 1, 7, 45, 0)}, {"start": datetime(2020, 1, 1, 17, 45, 0), "end": datetime(2020, 1, 2, 7, 45, 0)}]}}


        # constraint checker requests lower and upper bound, so we need some variants for various cases
        self.sunrise_data_early_night_early_night = {
            'CS001': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}, {"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}, {"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)},{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)}],
                      "night": [{"start": datetime(2019, 12, 31, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}, {"start": datetime(2019, 12, 31, 17, 30, 0), "end": datetime(2020, 1, 1, 7, 30, 0)}]}}

        self.sunrise_data_early_night_late_night = {
            'CS001': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}, {"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}, {"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)},{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)}],
                      "night": [{"start": datetime(2019, 12, 31, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}, {"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}]}}

        self.sunrise_data_late_night_late_night = {
            'CS001': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}, {"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}, {"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)},{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)}],
                      "night": [{"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}, {"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}]}}

        self.sunrise_data_late_night_early_night_next_day = {
            'CS001': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}, {"start": datetime(2020, 1, 2, 7, 30, 0), "end": datetime(2020, 1, 2, 9, 30, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}, {"start": datetime(2020, 1, 2, 9, 30, 0), "end": datetime(2020, 1, 2, 15, 30, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)},{"start": datetime(2020, 1, 2, 15, 30, 0), "end": datetime(2020, 1, 2, 17, 30, 0)}],
                      "night": [{"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}, {"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}]}}

        self.sunrise_data_late_night_late_night_next_day = {
            'CS001': {"sunrise": [{"start": datetime(2020, 1, 1, 7, 30, 0), "end": datetime(2020, 1, 1, 9, 30, 0)}, {"start": datetime(2020, 1, 2, 7, 30, 0), "end": datetime(2020, 1, 2, 9, 30, 0)}],
                      "day": [{"start": datetime(2020, 1, 1, 9, 30, 0), "end": datetime(2020, 1, 1, 15, 30, 0)}, {"start": datetime(2020, 1, 2, 9, 30, 0), "end": datetime(2020, 1, 2, 15, 30, 0)}],
                      "sunset": [{"start": datetime(2020, 1, 1, 15, 30, 0), "end": datetime(2020, 1, 1, 17, 30, 0)},{"start": datetime(2020, 1, 2, 15, 30, 0), "end": datetime(2020, 1, 2, 17, 30, 0)}],
                      "night": [{"start": datetime(2020, 1, 1, 17, 30, 0), "end": datetime(2020, 1, 2, 7, 30, 0)}, {"start": datetime(2020, 1, 2, 17, 30, 0), "end": datetime(2020, 1, 3, 7, 30, 0)}]}}


        self.sunrise_patcher = mock.patch('lofar.sas.tmss.services.scheduling.constraints.template_constraints_v1.timestamps_and_stations_to_sun_rise_and_set')
        self.sunrise_mock = self.sunrise_patcher.start()
        self.sunrise_mock.return_value = self.sunrise_data
        self.addCleanup(self.sunrise_patcher.stop)

    # require_day

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_day_start(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()
        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 4, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['day'][0]['start'])

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_day_start_of_latest_station(self):
        self.scheduling_unit_blueprint.requirements_doc['tasks']['Observation']['specifications_doc']['station_groups'] = [{'stations': ['CS001', 'DE601']}]
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()
        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 4, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['DE601']['day'][0]['start'])

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_timestamp(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, timestamp)

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_next_day_start(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 20, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['day'][1]['start'])

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_next_day_start_when_obs_does_not_fit(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['day'][1]['start'])

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_daytime_constraint_returns_true(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 10, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_daytime_constraint_returns_false_when_not_daytime(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 20, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_daytime_constraint_returns_false_when_partially_not_daytime(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 14, 0, 0)
        upper_bound = datetime(2020, 1, 1, 18, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 12, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daytime_constraint_returns_correct_value(self):
        # todo: for time ranges across dates, consider removing the mock for this because the moving window cannot be easily mocked
        # remove other constraints:
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['sky'] = {}

        # set constraint to test
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_day'] = True
        self.scheduling_unit_blueprint.save()

        # can run in day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(can_run_within_timewindow(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # cannot run at night
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 15, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertFalse(can_run_within_timewindow(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    # require_night

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_night_start(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['night'][0]['start'])

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_night_start_of_latest_station(self):
        self.scheduling_unit_blueprint.requirements_doc['tasks']['Observation']['specifications_doc']['station_groups'] = [{'stations': ['CS001', 'DE601']}]
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['DE601']['night'][0]['start'])

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_timestamp(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()

        # late night
        timestamp = datetime(2020, 1, 1, 23, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, timestamp)

        # early night
        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 3, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, timestamp)

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_next_night_start_when_obs_does_not_fit(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()

        # early night
        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 6, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data_early_night['CS001']['night'][1]['start'])

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_nighttime_constraint_returns_true(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()

        # early night
        self.sunrise_mock.return_value = self.sunrise_data_early_night_early_night
        lower_bound = datetime(2020, 1, 1, 1, 0, 0)
        upper_bound = datetime(2020, 1, 1, 3, 0, 0)
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # late night
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 20, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # night-night next day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_early_night_next_day
        lower_bound = datetime(2020, 1, 1, 23, 0, 0)
        upper_bound = datetime(2020, 1, 2, 3, 0, 0)
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_nighttime_constraint_returns_false_when_not_nighttime(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 10, 0, 0)
        upper_bound = datetime(2020, 1, 1, 14, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_nighttime_constraint_returns_false_when_partially_not_nighttime(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()

        # night-day next day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night_next_day
        lower_bound = datetime(2020, 1, 1, 23, 0, 0)
        upper_bound = datetime(2020, 1, 2, 10, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # day-night next day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_early_night_next_day
        lower_bound = datetime(2020, 1, 1, 14, 0, 0)
        upper_bound = datetime(2020, 1, 2, 3, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # day-night same day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 14, 0, 0)
        upper_bound = datetime(2020, 1, 1, 20, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # night-day same day
        self.sunrise_mock.return_value = self.sunrise_data_early_night_late_night
        lower_bound = datetime(2020, 1, 1, 3, 0, 0)
        upper_bound = datetime(2020, 1, 1, 10, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # day-night-day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night_next_day
        lower_bound = datetime(2020, 1, 1, 14, 0, 0)
        upper_bound = datetime(2020, 1, 2, 10, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # night-day-night
        self.sunrise_mock.return_value = self.sunrise_data_early_night_late_night
        lower_bound = datetime(2020, 1, 1, 3, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_nighttime_constraint_returns_correct_value(self):
        # todo: for time ranges across dates, consider removing the mock for this because the moving window cannot be easily mocked
        # remove other constraints:
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['sky'] = {}

        # set constraint to test
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['require_night'] = True
        self.scheduling_unit_blueprint.save()

        # cannot run in day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertFalse(can_run_within_timewindow(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # can run at night
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 15, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertTrue(can_run_within_timewindow(self.scheduling_unit_blueprint, lower_bound, upper_bound))


    # avoid_twilight

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_day_start(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 9, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['day'][0]['start'])

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_day_start_of_latest_station(self):
        self.scheduling_unit_blueprint.requirements_doc['tasks']['Observation']['specifications_doc']['station_groups'] = [{'stations': ['CS001', 'DE601']}]
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 9, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['DE601']['day'][0]['start'])

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_night_start(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data
        timestamp = datetime(2020, 1, 1, 17, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['night'][0]['start'])

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_night_start_of_latest_station(self):
        self.scheduling_unit_blueprint.requirements_doc['tasks']['Observation']['specifications_doc']['station_groups'] = [{'stations': ['CS001', 'DE601']}]
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data
        timestamp = datetime(2020, 1, 1, 17, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['DE601']['night'][0]['start'])

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_timestamp(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        # daytime
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, timestamp)

        # late time
        timestamp = datetime(2020, 1, 1, 20, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, timestamp)

        # early night
        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 3, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, timestamp)

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_day_or_night_start_when_obs_does_not_fit(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        timestamp = datetime(2020, 1, 1, 15, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['night'][0]['start'])

        self.sunrise_mock.return_value = self.sunrise_data_early_night
        timestamp = datetime(2020, 1, 1, 7, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint, timestamp)
        self.assertEqual(returned_time, self.sunrise_data['CS001']['day'][0]['start'])

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_twilight_constraint_returns_true(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 10, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_twilight_constraint_returns_false_when_in_twilight(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 9, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 16, 0, 0)
        upper_bound = datetime(2020, 1, 1, 17, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_anywhere_within_timewindow_with_daily_constraints_with_twilight_constraint_returns_false_when_partially_in_twilight(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 10, 0, 0)
        upper_bound = datetime(2020, 1, 1, 18, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 10, 0, 0)
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_daily_constraints(self.scheduling_unit_blueprint, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_twilight_constraint_returns_correct_value(self):
        # todo: for time ranges across dates, consider removing the mock for this because the moving window cannot be easily mocked
        # remove other constraints:
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['sky'] = {}

        # set constraint to test
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['daily']['avoid_twilight'] = True
        self.scheduling_unit_blueprint.save()

        # can run in day
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(can_run_within_timewindow(self.scheduling_unit_blueprint, lower_bound, upper_bound))

        # can run at night
        self.sunrise_mock.return_value = self.sunrise_data_late_night_late_night
        lower_bound = datetime(2020, 1, 1, 15, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertTrue(can_run_within_timewindow(self.scheduling_unit_blueprint, lower_bound, upper_bound))


class TestSkyConstraints(unittest.TestCase):
    '''
    Tests for the constraint checkers used in dynamic scheduling
    '''

    def setUp(self) -> None:
        # scheduling unit
        self.obs_duration = 120 * 60
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_unit_draft = TestDynamicScheduling.create_simple_observation_scheduling_unit("scheduling unit for ...%s" % self._testMethodName[30:],
                                                                                                scheduling_set=scheduling_set,
                                                                                                obs_duration=self.obs_duration)
        self.scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # mock out conversions for speedup and assertable timestamps
        self.distance_data = {
           "sun": {datetime(2020, 1, 1, 10, 0, 0): Angle("0.3rad"), datetime(2020, 1, 1, 12, 0, 0): Angle("0.35rad")},
           "moon": {datetime(2020, 1, 1, 10, 0, 0): Angle("0.2rad"), datetime(2020, 1, 1, 12, 0, 0): Angle("0.25rad")},
           "jupiter": {datetime(2020, 1, 1, 10, 0, 0): Angle("0.1rad"), datetime(2020, 1, 1, 12, 0, 0): Angle("0.15rad")}
        }
        self.distance_patcher = mock.patch('lofar.sas.tmss.services.scheduling.constraints.template_constraints_v1.coordinates_and_timestamps_to_separation_from_bodies')
        self.distance_mock = self.distance_patcher.start()
        self.distance_mock.return_value = self.distance_data
        self.addCleanup(self.distance_patcher.stop)
        
        self.target_rise_and_set_data = {"CS002": [{"rise": datetime(2020, 1, 1, 8, 0, 0), "set": datetime(2020, 1, 1, 12, 30, 0)},
                                                   {"rise": datetime(2020, 1, 1, 8, 0, 0), "set": datetime(2020, 1, 1, 12, 30, 0)}]}
        self.target_rise_and_set_patcher = mock.patch('lofar.sas.tmss.services.scheduling.constraints.template_constraints_v1.coordinates_timestamps_and_stations_to_target_rise_and_set')
        self.target_rise_and_set_mock = self.target_rise_and_set_patcher.start()
        self.target_rise_and_set_mock.return_value = self.target_rise_and_set_data
        self.addCleanup(self.target_rise_and_set_patcher.stop)

    # min_distance

    def test_can_run_anywhere_within_timewindow_with_sky_constraints_with_min_distance_constraint_returns_true_when_met(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['sky'] = {'min_distance': {'sun': 0.1, 'moon': 0.1, 'jupiter': 0.1}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_value = tc1.can_run_anywhere_within_timewindow_with_sky_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + timedelta(seconds=self.obs_duration))
        self.assertTrue(returned_value)

    def test_can_run_anywhere_within_timewindow_with_sky_constraints_with_min_distance_constraint_returns_false_when_not_met(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['sky'] = {'min_distance': {'sun': 0.2, 'moon': 0.2, 'jupiter': 0.2}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_value = tc1.can_run_anywhere_within_timewindow_with_sky_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + timedelta(seconds=self.obs_duration))
        self.assertFalse(returned_value)
        
    # min_target_elevation

    def test_can_run_anywhere_within_timewindow_with_sky_constraints_with_min_target_elevation_constraint_returns_true_when_met(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['sky'] = {'min_target_elevation': 0.1}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_value = tc1.can_run_anywhere_within_timewindow_with_sky_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + timedelta(seconds=self.obs_duration))
        self.assertTrue(returned_value)

    def test_can_run_anywhere_within_timewindow_with_sky_constraints_with_min_target_elevation_constraint_returns_false_when_not_met(self):
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['sky'] = {'min_target_elevation': 0.2}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 11, 0, 0)
        returned_value = tc1.can_run_anywhere_within_timewindow_with_sky_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + timedelta(seconds=self.obs_duration))
        self.assertFalse(returned_value)


class TestTimeConstraints(TestCase):
    """
    Tests for the time constraint checkers used in dynamic scheduling with different boundaries
    Possible time constraints are
    - after
    - before
    - between (one or more 'from-to')
    - not between (one or more 'from-to')
    """

    def add_time_between_constraint(self, from_timestamp, to_timestamp):
        lst_between_constraints = self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["between"]
        time_constraint_dict = {"from": from_timestamp.isoformat(), "to": to_timestamp.isoformat()}
        lst_between_constraints.append(time_constraint_dict)
        self.scheduling_unit_blueprint.save()

    def add_time_not_between_constraint(self, from_timestamp, to_timestamp):
        lst_between_constraints = self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["not_between"]
        time_constraint_dict = {"from": from_timestamp.isoformat(), "to": to_timestamp.isoformat()}
        lst_between_constraints.append(time_constraint_dict)
        self.scheduling_unit_blueprint.save()

    def setUp(self) -> None:
        # scheduling unit
        self.obs_duration = 120 * 60
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_unit_draft = TestDynamicScheduling.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=scheduling_set,
                                    obs_duration=self.obs_duration)
        self.scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

    def test_can_run_anywhere_after_returns_false(self):
        # Set datetime constraints after lower_bound
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 13, 0, 0).isoformat()
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                             datetime(2020, 1, 1, 12, 0, 0),
                                                                             datetime(2020, 1, 2, 12, 0, 0)))

        # Set datetime constraints to upper_bound
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 2, 12, 0, 0).isoformat()
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                             datetime(2020, 1, 1, 12, 0, 0),
                                                                             datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_anywhere_after_returns_true(self):
        # Set datetime constraints before lower_bound
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 11, 0, 0).isoformat()
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))
        # Set datetime constraints equal to lower_bound
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 12, 0, 0).isoformat()
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_anywhere_before_returns_false(self):
        # Set datetime constraints after upper_bound
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 2, 13, 0, 0).isoformat()
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                             datetime(2020, 1, 1, 12, 0, 0),
                                                                             datetime(2020, 1, 2, 12, 0, 0)))
        # Set datetime constraints equal to upper_bound
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 2, 12, 0, 0).isoformat()
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                             datetime(2020, 1, 1, 12, 0, 0),
                                                                             datetime(2020, 1, 2, 12, 0, 0)))
        # Set datetime constraints equal to upper_bound - duration + 1 sec
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["before"] = \
            (datetime(2020, 1, 2, 12, 0, 0) - self.scheduling_unit_blueprint.duration + timedelta(seconds=1)).isoformat()
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_anywhere_before_returns_true(self):
        # Set datetime constraints far before upper_bound (lower_bound)
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 1, 12, 0, 0).isoformat()
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))
        # Set datetime constraints equal to upper_bound - duration
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["before"] = \
            (datetime(2020, 1, 2, 12, 0, 0) - self.scheduling_unit_blueprint.duration).isoformat()
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_anywhere_between_returns_false(self):
        """
        Test 'between' constraint with start/stop datetime constraints 'outside' upper_bound or lower_bound
        """
        # Set datetime constraints start > lower_bound and stop > upper_bound
        self.add_time_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 2, 15, 0, 0))
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))
        # Set datetime constraints start < lower_bound and stop < upper_bound
        self.add_time_between_constraint(datetime(2020, 1, 1, 8, 0, 0), datetime(2020, 1, 2, 8, 0, 0))
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))
        # Set datetime constraints start > lower_bound and stop > upper_bound (1 second only)
        self.add_time_between_constraint(datetime(2020, 1, 1, 12, 0, 1), datetime(2020, 1, 2, 12, 0, 1))
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_anywhere_between_returns_true(self):
        """
        Test 'between' constraint with start/stop datetime constraints 'inside' upper_bound and lower_bound
        """
        # Set datetime constraints start > lower_bound and stop < upper_bound -duration
        self.add_time_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 15, 0, 0))
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 20, 0, 0)))

        # Set datetime constraints start = lower_bound and stop = upper_bound - duration
        self.add_time_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 15, 0, 0))
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 17, 10, 0)))

    def test_can_run_anywhere_not_between_returns_false(self):
        """
        Test 'not_between' constraint with start/stop datetime constraints 'inside' upper_bound or lower_bound
        """
        # Set datetime constraints start > lower_bound and stop > upper_bound
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 2, 15, 0, 0))
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                             datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))

        # Set datetime constraints start < lower_bound and stop > lower_bound and < upper_bound
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 8, 0, 0), datetime(2020, 1, 2, 8, 0, 0))
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                             datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))

        # Set datetime constraints start > lower_bound and stop < upper_bound
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 16, 0, 0), datetime(2020, 1, 2, 8, 0, 0))
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                             datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))

        # Set datetime constraints start < lower_bound and stop > upper_bound
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 8, 0, 0), datetime(2020, 1, 2, 14, 0, 0))
        self.assertFalse(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                            datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_anywhere_not_between_returns_true(self):
        """
        Test 'not_between' constraint with start/stop datetime constraints 'outside' upper_bound and lower_bound
        """
        # Set datetime constraints start < lower_bound and stop < lower_bound
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 3, 0, 0), datetime(2020, 1, 1, 11, 0, 0))
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                             datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 16, 0, 0)))

        # Set datetime constraints start > upper_bound and stop > upper_bound
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 16, 0, 0), datetime(2020, 1, 1, 20, 0, 0))
        self.assertTrue(tc1.can_run_anywhere_within_timewindow_with_time_constraints(self.scheduling_unit_blueprint,
                                             datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 15, 0, 0)))

    def execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary(self):
        """
        Just a simple wrapper to call 'can_run_anywhere_within_timewindow_with_time_constraints' function
        with a 24 hours boundary 2020-01-01 12:00 - 2020-01-02 12:00
        """
        return (tc1.can_run_within_timewindow_with_time_constraints(
                    self.scheduling_unit_blueprint, datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_within_between_constraints(self):
        """
        Test multiple 'between' constraints within 24 boundary and check overall result of
        'can_run_within_timewindow_with_time_constraints'
        This function will iterate between the boundary with boundary shift of 1hr and boundary length of
        the observation duration, which is in this testcase two hour
        i.e. 12-14, 13-15, 14-16,..etc.., 9-11
        """
        # no constraints defined so should be OK
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraints of 1hr, we still 'can_run'
        self.add_time_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 14, 0, 0))
        self.add_time_between_constraint(datetime(2020, 1, 1, 16, 0, 0), datetime(2020, 1, 1, 17, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraints of 2hr, we still 'can_run'
        self.add_time_between_constraint(datetime(2020, 1, 2, 11, 0, 0), datetime(2020, 1, 2, 13, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraint of 24hr constraint, we still 'can_run'
        self.add_time_between_constraint(datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraint of 2hr, to fill the 'last gap', we 'can run'
        self.add_time_between_constraint(datetime(2020, 1, 2, 10, 0, 0), datetime(2020, 1, 2, 12, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Clear all between constraints
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["between"] = []

        # Add constraints 'outside' the 24hr, now we 'can not run'
        self.add_time_between_constraint(datetime(2020, 1, 2, 13, 0, 0), datetime(2020, 1, 2, 14, 0, 0))
        self.add_time_between_constraint(datetime(2020, 1, 2, 16, 0, 0), datetime(2020, 1, 2, 17, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraint 'outside' the 24hr, we 'still can not run'
        self.add_time_between_constraint(datetime(2020, 1, 1, 9, 0, 0), datetime(2020, 1, 1, 12, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # add one 'inside' constraint, 1 hour within block of 2 hour so overall must be ok
        self.add_time_between_constraint(datetime(2020, 1, 1, 13, 30, 0), datetime(2020, 1, 1, 14, 30, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

    def test_can_run_within_not_between_constraints(self):
        """
        Test multiple 'not_between' constraints within 24 boundary and check overall result of
        'can_run_within_timewindow_with_time_constraints'
        This function will iterate between the boundary with boundary shift of 1hr and boundary length of
        the observation duration, which is in this testcase two hour
        i.e. 12-14, 13-15, 14-16,..etc.., 9-11
        """
        # no constraints defined so should be OK
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraints of 1hr, we still 'can_run'
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 14, 0, 0))
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 16, 0, 0), datetime(2020, 1, 1, 17, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraints of 2hr, we still 'can_run'
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 18, 0, 0), datetime(2020, 1, 1, 20, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraint of 20hr constraint, we still 'can_run'
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 8, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Add constraint of 2hr, to fill the 'last gap', now we can not run
        self.add_time_not_between_constraint(datetime(2020, 1, 2, 10, 0, 0), datetime(2020, 1, 2, 12, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())

        # Clear all not_between constraints
        self.scheduling_unit_blueprint.draft.scheduling_constraints_doc['time']["not_between"] = []
        # Add 4 hr constraints within 24 hours boundary, we can run
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 16, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_with_time_constraints_of_24hour_boundary())


class TestReservedStations(unittest.TestCase):
    """
    Tests for the reserved stations used in dynamic scheduling
    Test with different boundaries of scheduling unit start and stop times
    Reservation 'visualized'
          @ = station reservation start_time, * = station reservation stop_time
               SUB start_time     SUB stop_time                 Expected Result
     1.                |     @ ......|...*                      can NOT run
     2.             @..|..*          |                          can NOT run
     3.                |  @.....*    |                          can NOT run
     4.             @..|.............|......*                   can NOT run
     5.       @......* |             |                          can run
     6.                |             |   @.....*                can run
    """

    @staticmethod
    def create_station_reservation(additional_name, lst_stations, start_time=datetime(2100, 1, 1, 0, 0, 0), duration=86400):
        """
        Create a station reservation with given list of stations, start_time and duration (optional)
        Default duration is 24 hours (defined in seconds)
        """
        reservation_template = models.ReservationTemplate.objects.get(name="resource reservation")
        reservation_template_spec = get_default_json_object_for_schema(reservation_template.schema)
        reservation_template_spec["resources"] = {"stations": lst_stations }
        res = models.Reservation.objects.create(name="Station Reservation %s" % additional_name,
                                   description="Station reservation for testing",
                                   specifications_template=reservation_template,
                                   specifications_doc=reservation_template_spec,
                                   start_time=start_time,
                                   duration=duration)
        return res

    def setUp(self) -> None:
        # scheduling unit
        self.obs_duration = 120 * 60 # 2 hours
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_unit_draft = TestDynamicScheduling.create_simple_observation_scheduling_unit(
            "scheduling unit for %s" % self._testMethodName,
            scheduling_set=scheduling_set,
            obs_duration=self.obs_duration)
        self.scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(
            scheduling_unit_draft)
        # wipe all reservations in between tests, so the tests don't influence each other
        for reservation in models.Reservation.objects.all():
            reservation.delete()

    def set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(self, station_reservation):
        """
        Set (1) reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        """
        station_reservation.start_time = self.scheduling_unit_blueprint.start_time + timedelta(minutes=5)
        reservation_stop_time = self.scheduling_unit_blueprint.stop_time + timedelta(minutes=5)
        station_reservation.duration = (reservation_stop_time - station_reservation.start_time).total_seconds()
        station_reservation.save()

    def set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(self, station_reservation):
        """
        Set (2) reservation start_time < SUB start_time and reservation stop_time < SUB stop_time
        """
        station_reservation.start_time = self.scheduling_unit_blueprint.start_time - timedelta(minutes=5)
        reservation_stop_time = self.scheduling_unit_blueprint.stop_time - timedelta(minutes=5)
        station_reservation.duration = (reservation_stop_time - station_reservation.start_time).total_seconds()
        station_reservation.save()

    def set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(self, station_reservation):
        """
        Set (3) reservation start_time > SUB start_time and reservation stop_time < SUB stop_time
        """
        station_reservation.start_time = self.scheduling_unit_blueprint.start_time + timedelta(minutes=5)
        reservation_stop_time = self.scheduling_unit_blueprint.stop_time - timedelta(minutes=5)
        station_reservation.duration = (reservation_stop_time - station_reservation.start_time).total_seconds()
        station_reservation.save()

    def set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(self, station_reservation):
        """
        Set (4) reservation start_time < SUB start_time and reservation stop_time > SUB stop_time
        """
        station_reservation.start_time = self.scheduling_unit_blueprint.start_time - timedelta(minutes=5)
        reservation_stop_time = self.scheduling_unit_blueprint.stop_time + timedelta(minutes=5)
        station_reservation.duration = (reservation_stop_time - station_reservation.start_time).total_seconds()
        station_reservation.save()

    def set_5_reservation_start_time_and_stop_time_lt_sub_start_time(self, station_reservation):
        """
        Set (5) reservation start_time and reservation stop_time < SUB start_time
        """
        station_reservation.start_time = self.scheduling_unit_blueprint.start_time - timedelta(minutes=60)
        reservation_stop_time = self.scheduling_unit_blueprint.start_time - timedelta(minutes=5)
        station_reservation.duration = (reservation_stop_time - station_reservation.start_time).total_seconds()
        station_reservation.save()

    def set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(self, station_reservation):
        """
        Set (6) reservation start_time and reservation stop_time > SUB stop_time
        """
        station_reservation.start_time = self.scheduling_unit_blueprint.stop_time + timedelta(minutes=5)
        reservation_stop_time = self.scheduling_unit_blueprint.stop_time + timedelta(minutes=65)
        station_reservation.duration = (reservation_stop_time - station_reservation.start_time).total_seconds()
        station_reservation.save()

    def update_station_groups_of_scheduling_unit_blueprint(self):
        """
        Use the UC1 strategy template to 'easily' extend the station group of the scheduling_unit
        For info, it will have three station groups
        - dutch station with max_nr_missing=4
        - international with max_nr_missing=2
        - international required with max_nr_missing=1
        """
        uc1_strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")
        scheduling_unit_spec = add_defaults_to_json_object_for_schema(uc1_strategy_template.template,
                                                                      uc1_strategy_template.scheduling_unit_template.schema)
        station_groups = scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['station_groups']
        self.scheduling_unit_blueprint.requirements_doc['tasks']['Observation']['specifications_doc']['station_groups'] = station_groups

    def test_one_station_reserved(self):
        """
        Test station reservation when 1 station (CS001) is reserved and station CS001 is used in scheduling_unit
        with different reservation start and stop times
        """
        reservation_one = self.create_station_reservation("One", ["CS001"])
        # reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        self.set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_one)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time < SUB start_time and stop_time < SUB stop_time
        self.set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_one)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time > SUB start_time and stop_time < SUB stop_time
        self.set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_one)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time < SUB start_time and stop_time > SUB stop_time
        self.set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_one)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # Reservations outside boundary
        # start_time and stop_time < SUB start_time
        self.set_5_reservation_start_time_and_stop_time_lt_sub_start_time(reservation_one)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # start_time and stop_time > SUB stop_time
        self.set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(reservation_one)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_two_stations_reserved(self):
        """
        Test station reservation when 2 station (CS001,CS002) are reserved and station CS001 is used in scheduling_unit
        with different reservation start and stop times
        """
        reservation_two = self.create_station_reservation("Two", ["CS001", "CS002"])
        # reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        self.set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time < SUB start_time and stop_time < SUB stop_time
        self.set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time > SUB start_time and stop_time < SUB stop_time
        self.set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time < SUB start_time and stop_time > SUB stop_time
        self.set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # Reservations outside boundary
        # start_time and stop_time < SUB start_time
        self.set_5_reservation_start_time_and_stop_time_lt_sub_start_time(reservation_two)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # start_time and stop_time > SUB stop_time
        self.set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(reservation_two)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_two_stations_reserved_but_not_used(self):
        """
        Test station reservation when 2 stations (CS002, CS003) are reserved and station CS001 is used in scheduling_unit
        with different reservation start and stop times
        All possibilities should result in 'can run'
        """
        reservation_two_no_overlap = self.create_station_reservation("Two-NoOverlap", ["CS002", "CS003"])
        # reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        self.set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two_no_overlap)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time < SUB start_time and stop_time < SUB stop_time
        self.set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two_no_overlap)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time > SUB start_time and stop_time < SUB stop_time
        self.set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two_no_overlap)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time < SUB start_time and stop_time > SUB stop_time
        self.set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two_no_overlap)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # Reservations outside boundary
        # start_time and stop_time < SUB start_time
        self.set_5_reservation_start_time_and_stop_time_lt_sub_start_time(reservation_two_no_overlap)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # start_time and stop_time > SUB stop_time
        self.set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(reservation_two_no_overlap)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_two_stations_reserved_with_duration_null(self):
        """
        Test station reservation when two stations (CS001,CS002) are reserved with duration null and so reserved indefinitely
        and station CS001 is used in scheduling_unit
        Test with different reservation start time and NO stop_time
        start_time after SUB stop_time 'can run' all others 'can NOT run'
        """
        reservation_two_no_duration = self.create_station_reservation("Two-NoDuration", ["CS001", "CS002"], duration=None)
        # reservation start_time > SUB start_time and < SUB stop_time
        reservation_two_no_duration.start_time = self.scheduling_unit_blueprint.start_time + timedelta(minutes=5)
        reservation_two_no_duration.save()
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time < SUB start_time (and < SUB stop_time of course)
        reservation_two_no_duration.start_time = self.scheduling_unit_blueprint.start_time - timedelta(minutes=5)
        reservation_two_no_duration.save()
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

        # reservation start_time > SUB stop time
        reservation_two_no_duration.start_time = self.scheduling_unit_blueprint.stop_time + timedelta(minutes=5)
        reservation_two_no_duration.save()
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_dutch_stations_conflicts_result_false(self):
        """
        Test conflict of 'Dutch' station which have a default of max_nr_missing=4,
        Create stations reservation equal to max_nr_missing+1 and check that it can not run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Dutch", ['CS001', 'CS002', 'CS003', 'CS401', 'CS501'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_dutch_stations_conflicts_result_true(self):
        """
        Test conflict of 'Dutch' station which have a default of max_nr_missing=4,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Dutch", ['CS001', 'CS002', 'CS003', 'CS401'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_stations_conflicts_result_false(self):
        """
        Test conflict of 'International' stations which have a default of max_nr_missing=2,
        Create stations reservation equal to max_nr_missing+1 and check that it can not run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International", ['SE607', 'PL610', 'PL612'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_stations_conflicts_result_true(self):
        """
        Test conflict of 'International' stations which are have a default of max_nr_missing=2,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International", ['SE607', 'PL610'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_required_stations_conflicts_result_false(self):
        """
        Test conflict of 'International Required' stations which are have a default of max_nr_missing=1,
        Create stations reservation equal to max_nr_missing+1 and check that it can not run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International Required", ['DE601', 'DE605'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_required_stations_conflicts_result_true(self):
        """
        Test conflict of 'International Required' stations which are have a default of max_nr_missing=1,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International Required", ['DE605'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_mixed_required_stations_conflicts_result_false(self):
        """
        Test conflict of 'mixed' stations which are have a default of max_nr_missing,
        Create stations reservation equal to max_nr_missing and one station group max_nr_missing+1
        and check that it can not run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Mixed", ['DE605', 'SE607', 'PL610', 'CS001', 'CS002', 'CS003', 'CS401'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_mixed_required_stations_conflicts_result_true(self):
        """
        Test conflict of 'mixed' stations which are have a default of max_nr_missing,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        self.update_station_groups_of_scheduling_unit_blueprint()
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Mixed", ['DE605', 'PL610', 'CS001', 'CS002', 'CS003', 'CS401'],
                                        start_time=self.scheduling_unit_blueprint.start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))


logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

if __name__ == '__main__':
    #run the unit tests
    unittest.main()
