#!/usr/bin/env python3

# subtask_scheduling.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.


import logging
import os
from optparse import OptionParser, OptionGroup

logger = logging.getLogger(__name__)

from lofar.common import dbcredentials
from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.common.util import find_free_port

from enum import Enum
from json import dumps as JSONdumps
from SimpleWebSocketServer import SimpleWebSocketServer, WebSocket
from threading import Thread, Event
from django.apps import apps

DEFAULT_WEBSOCKET_PORT = 5678

'''
NOTE: 
HTTP Headers are not supported inside the WebSocket protocol (see https://tools.ietf.org/html/rfc6455#page-11).
WS traffic is either UTC-8 text or binary data (see https://tools.ietf.org/html/rfc6455#page-39).
'''

class TMSSEventMessageHandlerForWebsocket(TMSSEventMessageHandler):
    '''
    '''
    class ObjActions(Enum):
        CREATE = 'create'
        UPDATE = 'update'
        DELETE = 'delete'

    class ObjTypes(Enum):   # These values refer to the DRF APIs
        SCHED_UNIT_BLUEPRINT = 'scheduling_unit_blueprint'
        SCHED_UNIT_DRAFT = 'scheduling_unit_draft'
        SUBTASK = 'subtask'
        TASK_BLUEPRINT = 'task_blueprint'
        TASK_DRAFT = 'task_draft'

    def __init__(self, websocket_port: int=DEFAULT_WEBSOCKET_PORT):
        super().__init__(log_event_messages=True)
        self.websocket_port = websocket_port
        self._run_ws = True

    def start_handling(self):
        socket_started_event = Event()

        # Create and run a simple ws server
        def start_ws_server():
            logger.debug("starting websocket server on port %s", self.websocket_port)
            self._ws_server = SimpleWebSocketServer('', self.websocket_port, WebSocket)
            socket_started_event.set()
            logger.info("started websocket server on port %s", self.websocket_port)
            while self._run_ws: # Run the server till the stop_handling
                self._ws_server.serveonce()

        self.t = Thread(target=start_ws_server)
        self.t.start()
        if not socket_started_event.wait(10):
            raise RuntimeError("Could not start websocket server on port %s"%self.websocket_port)
        super().start_handling()

    def stop_handling(self):
        super().stop_handling()
        self._run_ws = False    # Stop the ws server
        self.t.join()

    def _broadcast_notify_websocket(self, msg):
        # Send a broadcast message to all connected ws clients
        for ws in self._ws_server.connections.values():
            ws.sendMessage(JSONdumps(msg))    # Stringify msg, so clients can parse it as JSON

    def _post_update_on_websocket(self, id, object_type, action):
        # Prepare the json_blob_template
        json_blob = {'object_details': {'id': id}, 'object_type': object_type.value, 'action': action.value}
        if action == self.ObjActions.CREATE or action == self.ObjActions.UPDATE:
            try:
                model_class = apps.get_model("tmssapp", object_type.value.replace('_',''))
                model_instance = model_class.objects.get(id=id)
                if hasattr(model_instance, 'start_time') and model_instance.start_time is not None:
                    json_blob['object_details']['start_time'] = model_instance.start_time.isoformat()
                if hasattr(model_instance, 'stop_time') and model_instance.stop_time is not None:
                    json_blob['object_details']['stop_time'] = model_instance.stop_time.isoformat()
                if hasattr(model_instance, 'duration') and model_instance.duration is not None:
                    json_blob['object_details']['duration'] = model_instance.duration.total_seconds()
                if hasattr(model_instance, 'status'):
                    json_blob['object_details']['status'] = model_instance.status
                if hasattr(model_instance, 'state'):
                    json_blob['object_details']['state'] = model_instance.state.value
            except Exception as e:
                logger.error("Cannot get object details for %s: %s", json_blob, e)

        # Send the json_blob as a broadcast message to all connected ws clients
        self._broadcast_notify_websocket(json_blob)

    def onSubTaskCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SUBTASK, self.ObjActions.CREATE)

    def onSubTaskUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SUBTASK, self.ObjActions.UPDATE)

    def onSubTaskDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SUBTASK, self.ObjActions.DELETE)

    def onTaskDraftCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_DRAFT, self.ObjActions.CREATE)

    def onTaskDraftUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_DRAFT, self.ObjActions.UPDATE)

    def onTaskDraftDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_DRAFT, self.ObjActions.DELETE)

    def onTaskBlueprintCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.CREATE)

    def onTaskBlueprintUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.UPDATE)

    def onTaskBlueprintDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.DELETE)

    def onSchedulingUnitDraftCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.CREATE)

    def onSchedulingUnitDraftUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.UPDATE)

    def onSchedulingUnitDraftDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.DELETE)

    def onSchedulingUnitBlueprintCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.CREATE)

    def onSchedulingUnitBlueprintUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.UPDATE)

    def onSchedulingUnitBlueprintDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.DELETE)

def create_service(websocket_port: int=DEFAULT_WEBSOCKET_PORT, exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER):
    return TMSSBusListener(handler_type=TMSSEventMessageHandlerForWebsocket,
                           handler_kwargs={'websocket_port': websocket_port},
                           exchange=exchange, broker=broker)


def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_websocket_service which listens for TMSS event messages on the messagebus, and posts the updates on the websocket for htpp clients.')

    parser.add_option('-w', '--websocket_port', dest='websocket_port', type='int', default=find_free_port(DEFAULT_WEBSOCKET_PORT),
                      help='The port on which the websocket clients can connect, default: %default')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="exchange where the TMSS event messages are published. [default: %default]")
    parser.add_option_group(group)

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')

    (options, args) = parser.parse_args()

    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    with create_service(options.websocket_port, options.exchange, options.broker):
        waitForInterrupt()

if __name__ == '__main__':
    main()
