#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import uuid

import logging
logger = logging.getLogger('lofar.' + __name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.sas.tmss.test.test_utils import TMSSTestEnvironment

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor
from lofar.sas.tmss.services.websocket_service import create_service, TMSSEventMessageHandlerForWebsocket, DEFAULT_WEBSOCKET_PORT
from lofar.common.test_utils import integration_test
from lofar.common.util import single_line_with_single_spaces, find_free_port

from collections import deque
from json import loads as JSONLoads
import requests
from threading import Thread, Event
import websocket


@integration_test
class TestSubtaskSchedulingService(unittest.TestCase):
    '''
    Tests for the SubtaskSchedulingService
    '''

    def __init__(self, methodName: str = ...) -> None:
        self.msg_queue = deque()
        self.sync_event = Event()
        self.ObjActions = TMSSEventMessageHandlerForWebsocket.ObjActions
        self.ObjTypes = TMSSEventMessageHandlerForWebsocket.ObjTypes
        super().__init__(methodName)

    def start_ws_client(self, websocket_port: int=DEFAULT_WEBSOCKET_PORT):
        # Setup and start the Websocket client

        def on_message(ws, message):
            json_doc = JSONLoads(message)
            logger.info('Received msg from ws: %s', single_line_with_single_spaces(json_doc))
            self.msg_queue.append(json_doc)
            self.sync_event.set()

        def on_error(ws, error):
            logger.info(error)

        def on_open(ws):
            logger.info('Connected to ws')

        def on_close(ws):
            logger.info('Closed ws')

        def thread_ws_starter():
            self.ws = websocket.WebSocketApp("ws://127.0.0.1:%d/"%(websocket_port,),
                                        on_open=on_open,
                                        on_message=on_message,
                                        on_error=on_error,
                                        on_close=on_close)
            self.ws.run_forever()

        self.t = Thread(target=thread_ws_starter, daemon=True)
        self.t.start()

    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = uuid.uuid1()

        cls.tmp_exchange = TemporaryExchange("%s_%s" % (cls.__name__, cls.TEST_UUID))
        cls.tmp_exchange.open()

        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address, populate_test_data=False, populate_schemas=False, start_websocket=False, start_postgres_listener=True, enable_viewflow=False)
        cls.tmss_test_env.start()

        cls.test_data_creator = cls.tmss_test_env.create_test_data_creator()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()

    def test_01(self):
        '''
        This test starts a websocket service and tmss. Creates, updates and deletes objects to check if json_blobs from the ws service are properly received.
        '''

        logger.info(' -- test_01_for_expected_behaviour -- ')

        websocket_port = find_free_port(DEFAULT_WEBSOCKET_PORT)

        # create and start the service (the object under test)
        service = create_service(websocket_port=websocket_port, exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):

            self.start_ws_client(websocket_port)  # Start ws client

            def test_object(json_test, obj_type, action):   # Check if the correct/expected json_blobs arrive in the ws client
                # Wait for incoming ws message
                if not self.sync_event.wait(timeout=5):
                    raise TimeoutError()
                self.sync_event.clear()
                # Assert json_blobs
                json_blob = {'object_details': {'id': json_test['id']}, 'object_type': obj_type.value, 'action': action.value}
                if action == self.ObjActions.CREATE or action == self.ObjActions.UPDATE:
                    for key in ('start_time', 'stop_time', 'duration', 'status'):
                        if json_test.get(key) is not None:
                            json_blob['object_details'][key] = json_test[key]
                    if json_test.get('state_value') is not None:
                        json_blob['object_details']['state'] = json_test['state_value']
                self.assertEqual(json_blob, self.msg_queue.popleft())

            # Test creations
            # Test scheduling_unit_draft create
            su_draft = self.test_data_creator.post_data_and_get_response_as_json_object(
                self.test_data_creator.SchedulingUnitDraft(), '/scheduling_unit_draft/')
            test_object(su_draft, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.CREATE)

            # Test task_draft create
            task_draft = self.test_data_creator.post_data_and_get_response_as_json_object(
                self.test_data_creator.TaskDraft(scheduling_unit_draft_url=su_draft['url']), '/task_draft/')
            test_object(task_draft, self.ObjTypes.TASK_DRAFT, self.ObjActions.CREATE)

            # Test scheduling_unit_blueprint create
            su_blueprint = self.test_data_creator.post_data_and_get_response_as_json_object(
                self.test_data_creator.SchedulingUnitBlueprint(scheduling_unit_draft_url=su_draft['url']),
                '/scheduling_unit_blueprint/')
            test_object(su_blueprint, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.CREATE)

            # Test task_blueprint create
            task_blueprint = self.test_data_creator.post_data_and_get_response_as_json_object(
                self.test_data_creator.TaskBlueprint(scheduling_unit_blueprint_url=su_blueprint['url'],
                                                     draft_url=task_draft['url']), '/task_blueprint/')
            test_object(task_blueprint, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.CREATE)

            # Test subtask create
            subtask = self.test_data_creator.post_data_and_get_response_as_json_object(
                self.test_data_creator.Subtask(task_blueprint_url=task_blueprint['url']), '/subtask/')
            test_object(subtask, self.ObjTypes.SUBTASK, self.ObjActions.CREATE)

            # Test updates
            with self.tmss_test_env.create_tmss_client() as client:
                # Test subtask update
                client.set_subtask_status(subtask['id'], 'scheduled')
                subtask = requests.get(subtask['url'], auth=self.test_data_creator.auth).json()
                test_object(subtask, self.ObjTypes.SUBTASK, self.ObjActions.UPDATE)

                # Test task_blueprint update
                # TODO: fix. These requests were broken by permissions.
                # Probably we want to only send the id and action, and let the client do a GET on the url, because the client can auth himself.
                # task_blueprint = requests.get(task_blueprint['url'], auth=self.test_data_creator.auth).json()
                # test_object(task_blueprint, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.UPDATE)

                # # Test scheduling_unit_blueprint update
                # su_blueprint = requests.get(su_blueprint['url'], auth=self.test_data_creator.auth).json()
                # test_object(su_blueprint, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.UPDATE)

                # # Test scheduling_unit_draft update
                # su_draft['description'] = 'This is an update test'
                # su_draft = requests.put(su_draft['url'], json=su_draft, auth=self.test_data_creator.auth).json()
                # test_object(su_draft, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.UPDATE)
                #
                # # Test task_draft update
                # task_draft['description'] = 'This is an update test'
                # task_draft = requests.put(task_draft['url'], json=task_draft, auth=self.test_data_creator.auth).json()
                # test_object(task_draft, self.ObjTypes.TASK_DRAFT, self.ObjActions.UPDATE)

            # # Test deletions
            # # Test subtask delete
            # requests.delete(subtask['url'], auth=self.test_data_creator.auth)
            # test_object({'id': subtask['id']}, self.ObjTypes.SUBTASK, self.ObjActions.DELETE)
            #
            # # Test task_blueprint delete
            # requests.delete(task_blueprint['url'], auth=self.test_data_creator.auth)
            # test_object({'id': task_blueprint['id']}, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.DELETE)
            #
            # # Test scheduling_unit_blueprint delete
            # requests.delete(su_blueprint['url'], auth=self.test_data_creator.auth)
            # test_object({'id': su_blueprint['id']}, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.DELETE)
            #
            # # Test task_draft delete
            # requests.delete(task_draft['url'], auth=self.test_data_creator.auth)
            # test_object({'id': task_draft['id']}, self.ObjTypes.TASK_DRAFT, self.ObjActions.DELETE)
            #
            # # Test scheduling_unit_draft delete
            # requests.delete(su_draft['url'], auth=self.test_data_creator.auth)
            # test_object({'id': su_draft['id']}, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.DELETE)


if __name__ == '__main__':
    #run the unit tests
    unittest.main()
