"""
Django settings for tmss project.

Generated by 'django-admin startproject' using Django 2.0.6.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.0/ref/settings/
"""

import os
import logging
from lofar.common import dbcredentials, isDevelopmentEnvironment, isTestEnvironment

# remove basic handler, and allow TMSS/Django to do more complicated logging setup
for h in logging.root.handlers[:]:
    logging.root.removeHandler(h)
    h.close()

logger = logging.getLogger(__name__)

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'lofar_formatter': {
            'format': '%(asctime)s %(levelname)s %(message)s',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'lofar_formatter',
        },
    },
    'loggers': {
        'django.server': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False,
        },
        'django_auth_ldap': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'django.request': {
            'handlers': ['console'],
            'level': 'INFO',  # change debug level as appropiate
            'propagate': False,
        },
        # 'django.db.backends': { # uncomment to enable logging of each db query. Very spammy and slow, but also usefull for performance improvement. Gives more even detail/insight than django debug toolbar.
        #     'level': 'DEBUG',
        #     'handlers': ['console'],
        #     'propagate': False,
        # },
        'lofar': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False,
        },
    }
}

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.getenv('SECRET_KEY', '-&$!kx$_0)u1x#zk9w^^81hfssaover2(8wdq_8n8n3u(8=-9n')       # todo: set something new here for production !!!

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = bool(int(os.getenv('DEBUG', True)))

ALLOWED_HOSTS = os.getenv('ALLOWED_HOSTS').split(',') if os.getenv('ALLOWED_HOSTS') else []

# Application definition

INSTALLED_APPS = [
    'lofar.sas.tmss.tmss.tmssapp',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'rest_framework.authtoken',
    'django_jsonforms',
    'django_json_widget',
    'jsoneditor',
    'drf_yasg',
    'django_filters',
    'django_property_filter'
    ]

try:
    # material is nice but not mandatory
    import material
    INSTALLED_APPS.append('material')

    import material.frontend
    INSTALLED_APPS.append('material.frontend')
except ImportError as e:
    logger.warning(str(e))

MIDDLEWARE = [
    'django.middleware.gzip.GZipMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware'
]

def show_debug_toolbar(*args, **kwargs):
    return bool(os.environ.get('SHOW_DJANGO_DEBUG_TOOLBAR', False))

if show_debug_toolbar():
    DEBUG_TOOLBAR_CONFIG = { 'SHOW_TOOLBAR_CALLBACK': show_debug_toolbar }
    INSTALLED_APPS.append('debug_toolbar')
    MIDDLEWARE.insert(MIDDLEWARE.index('django.middleware.gzip.GZipMiddleware')+1, 'debug_toolbar.middleware.DebugToolbarMiddleware')

if os.environ.get('TMSS_ENABLE_VIEWFLOW', "False").lower() == "true":
    INSTALLED_APPS.extend(['viewflow', 'viewflow.frontend', 'lofar.sas.tmss.tmss.workflowapp'])


ROOT_URLCONF = 'lofar.sas.tmss.tmss.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [BASE_DIR, os.path.join(BASE_DIR, 'templates'), os.path.join(BASE_DIR, 'frontend','tmss_webapp')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

STATICFILES_DIRS = [
    os.path.join(BASE_DIR, 'frontend','tmss_webapp/build/static')
]

WSGI_APPLICATION = 'lofar.sas.tmss.tmss.wsgi.application'

# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases

# Read Postgres DB settings

# this requires a config file in ~/.lofar/dbcredentials/tmss.ini
# if no such file exists then it is automatically created.
# It is also possible to set another credentials-name via the TMSS_DBCREDENTIALS env-var for testing for example.
#
# contents should be like this (adapt as needed):
#
# [database:tmss]
# type = postgresql
# host = localhost
# database = <your_tmss_database_name>
# port = 5432
# user = <your_tmss_user_name>
# password = <your_tmss_password>
DATABASES = {
    'default': {
        # Postgres:
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': '',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}

if "POSTGRES_HOST" in  os.environ.keys():
    DATABASES['default']['HOST'] = os.getenv('POSTGRES_HOST')
    DATABASES['default']['PORT'] = os.getenv('POSTGRES_PORT')
    DATABASES['default']['NAME'] = os.getenv('POSTGRES_DB')
    DATABASES['default']['USER'] = os.getenv('POSTGRES_USER')
    DATABASES['default']['PASSWORD'] = os.getenv('POSTGRES_PASSWORD')
else:
    from lofar.common import dbcredentials, isDevelopmentEnvironment

    creds_name = os.environ.get('TMSS_DBCREDENTIALS', 'TMSS')
    django_db_credentials = dbcredentials.DBCredentials().get(creds_name)
    logger.debug("TMSS Django settings: Using dbcreds '%s' for django database: %s",
                 creds_name, django_db_credentials.stringWithHiddenPassword())
    DATABASES['default']['HOST'] = django_db_credentials.host
    DATABASES['default']['PORT'] = django_db_credentials.port
    DATABASES['default']['NAME'] = django_db_credentials.database
    DATABASES['default']['USER'] = django_db_credentials.user
    DATABASES['default']['PASSWORD'] = django_db_credentials.password

REST_FRAMEWORK = {
    # Authentication, see also if-tree below for Open-ID vs LDAP configuration
    'DEFAULT_AUTHENTICATION_CLASSES': [],
    'DEFAULT_PERMISSION_CLASSES': [],
    'PAGE_SIZE': 50,
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'DEFAULT_FILTER_BACKENDS': (
        # allows ?field=value filtering in URLs
        # (fields allowed for filtering must be annotated in viewsets using filter_fields = ('field', ...))
        'django_filters.rest_framework.DjangoFilterBackend',

        # allows ?ordering=[-]field to order list results
        'rest_framework.filters.OrderingFilter',
    ),
}


# AUTHENTICATION: simple LDAP, or OpenID, or both

AUTHENTICATION_BACKENDS = ('django.contrib.auth.backends.ModelBackend',)

if "TMSS_LDAPCREDENTIALS" in os.environ.keys():
    # plain LDAP
    import ldap

    REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'].append('rest_framework.authentication.BasicAuthentication')
    REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'].append('rest_framework.authentication.SessionAuthentication')
    REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'].append('rest_framework.authentication.TokenAuthentication')
    REST_FRAMEWORK['DEFAULT_PERMISSION_CLASSES'].append('rest_framework.permissions.IsAuthenticated')

    # LDAP
    ldap_creds_name = os.environ.get('TMSS_LDAPCREDENTIALS', 'TMSS_LDAP')
    django_ldap_credentials = dbcredentials.DBCredentials().get(ldap_creds_name)
    logger.info("TMSS Django settings: Using dbcreds '%s' for ldap authentication: %s",
                ldap_creds_name, django_ldap_credentials.stringWithHiddenPassword())

    AUTH_LDAP_GLOBAL_OPTIONS = {ldap.OPT_X_TLS_REQUIRE_CERT: ldap.OPT_X_TLS_NEVER}  # cert still expired?
    AUTH_LDAP_CONNECTION_OPTIONS = {ldap.OPT_X_TLS_REQUIRE_CERT: ldap.OPT_X_TLS_NEVER}  # cert still expired?
    protocol = 'ldap://' if isDevelopmentEnvironment() or isTestEnvironment() else 'ldaps://'
    AUTH_LDAP_SERVER_URI = "%s%s:%s" % (protocol, django_ldap_credentials.host, django_ldap_credentials.port)
    AUTH_LDAP_USER_DN_TEMPLATE = "cn=%(user)s,ou=Users,o=lofar,c=eu"

    AUTH_LDAP_USER_ATTR_MAP = {
        "first_name": "givenName",
        "last_name": "sn",
        "email": "mail"
    }

    AUTHENTICATION_BACKENDS += ('django_auth_ldap.backend.LDAPBackend',)


if "OIDC_RP_CLIENT_ID" in os.environ.keys():

    INSTALLED_APPS.append('mozilla_django_oidc')  # Load after auth
    REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'].append('mozilla_django_oidc.contrib.drf.OIDCAuthentication')
    REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'].append('rest_framework.authentication.SessionAuthentication')
    REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'].append('rest_framework.authentication.TokenAuthentication')
    REST_FRAMEWORK['DEFAULT_PERMISSION_CLASSES'].append('rest_framework.permissions.IsAuthenticated')

    # OPEN-ID CONNECT
    OIDC_DRF_AUTH_BACKEND = 'mozilla_django_oidc.auth.OIDCAuthenticationBackend'
    # For talking to Mozilla Identity Provider:
    OIDC_RP_SCOPES = "openid email profile"  # todo: groups are not a standard scope, how to handle those?
    OIDC_RP_CLIENT_ID = os.environ.get('OIDC_RP_CLIENT_ID', '2')  # Secret, do not put real credentials on Git
    OIDC_RP_CLIENT_SECRET = os.environ.get('OIDC_RP_CLIENT_SECRET',
                                           'secret')  # Secret, do not put real credentials on Git
    OIDC_ENDPOINT_HOST = os.environ.get('OIDC_ENDPOINT_HOST', 'localhost')
    OIDC_OP_AUTHORIZATION_ENDPOINT = os.environ.get('OIDC_OP_AUTHORIZATION_ENDPOINT', "http://localhost:8088/openid/authorize/")
    OIDC_OP_TOKEN_ENDPOINT = os.environ.get('OIDC_OP_TOKEN_ENDPOINT', "http://localhost:8088/openid/token/")
    OIDC_OP_USER_ENDPOINT = os.environ.get('OIDC_OP_USER_ENDPOINT', "http://localhost:8088/openid/userinfo/")

    AUTHENTICATION_BACKENDS += ('mozilla_django_oidc.auth.OIDCAuthenticationBackend',)
    MIDDLEWARE.append('mozilla_django_oidc.middleware.SessionRefresh')

if len(AUTHENTICATION_BACKENDS) == 1:
    REST_FRAMEWORK['DEFAULT_PERMISSION_CLASSES'].append('rest_framework.permissions.AllowAny')  # todo: Whoo! This seems unsafe! Maybe we should at least have users explicitly disable authentication on startup?!
    logger.warning("No authentication configured! please set either OIDC_RP_CLIENT_ID or TMSS_LDAPCREDENTIALS environment variable.")

LOGIN_REDIRECT_URL = "/api/"
LOGIN_REDIRECT_URL_FAILURE = "/api/"
LOGOUT_REDIRECT_URL = "/api/"
LOGOUT_REDIRECT_URL_FAILURE = "/api/"


# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/

LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = False  # We don't want timezone support since everything is UTC anyway and this caused trouble in the past


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.0/howto/static-files/
STATIC_URL = '/static/'
STATIC_ROOT = '/staticfiles/'



# Setup support for proxy 
USE_X_FORWARDED_HOST = True
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

SWAGGER_SETTINGS = {

    'SECURITY_DEFINITIONS': {
        'basic': {
            'type': 'basic'
        }
    },

}
