#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
import requests

import logging
logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
rest_data_creator = TMSSRESTTestDataCreator(BASE_URL, AUTH)

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.exceptions import SubtaskInvalidStateException
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset, convert_to_parset_dict
from lofar.common.json_utils import get_default_json_object_for_schema
from lofar.sas.tmss.tmss.tmssapp.adapters.sip import generate_sip_for_dataproduct
from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import append_to_subtask_raw_feedback, process_feedback_into_subtask_dataproducts, process_feedback_for_subtask_and_set_to_finished_if_complete, reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete
from lofar.lta.sip import constants
from lofar.parameterset import parameterset

from lofar.sas.resourceassignment.resourceassignmentestimator.resource_estimators import ObservationResourceEstimator


class ObservationParsetAdapterTest(unittest.TestCase):
    def get_default_specifications(self):
        subtask_template = models.SubtaskTemplate.objects.get(name='observation control')
        return get_default_json_object_for_schema(subtask_template.schema)

    def create_subtask(self, specifications_doc):
        subtask_template = models.SubtaskTemplate.objects.get(name='observation control')
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc)
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))
        dataproduct:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output))
        return subtask

    def test_correlator(self):
        specifications_doc = self.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = True
        specifications_doc['stations']['digital_pointings'] = [
          { "name": "target1",
            "subbands": list(range(8))
          }
        ]

        nr_files = 8 # = nr of subbands

        subtask = self.create_subtask(specifications_doc)
        parset = convert_to_parset_dict(subtask)
        logger.info("test_correlator parset:",parset)

        self.assertEqual(True, parset["Observation.DataProducts.Output_Correlated.enabled"])
        self.assertEqual(False, parset["Observation.DataProducts.Output_CoherentStokes.enabled"])
        self.assertEqual(False, parset["Observation.DataProducts.Output_IncoherentStokes.enabled"])
        self.assertEqual(False, parset["Cobalt.BeamFormer.flysEye"])

        # check whether parset is accepted by the ResourceEstimator
        estimator = ObservationResourceEstimator()
        estimations = estimator.verify_and_estimate(convert_to_parset_dict(subtask))
        self.assertEqual([],       estimations["errors"])

        # check whether the ResourceEstimator agrees with our spec
        self.assertEqual(nr_files, estimations["estimates"][0]["output_files"]["uv"][0]["properties"]["nr_of_uv_files"] * estimations["estimates"][0]["resource_count"])

    def test_flyseye(self):
        specifications_doc = self.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = False
        specifications_doc['stations']['station_list'] = ['CS001', 'CS002', 'RS205']
        specifications_doc['stations']['antenna_set'] = 'HBA_DUAL'
        specifications_doc['stations']['digital_pointings'] = [
          { "name": "target1",
            "subbands": list(range(8))
          }
        ]

        specifications_doc['COBALT']['beamformer']['flyseye_pipelines'] = [
            { "coherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              }
            }
        ]

        nr_files = 5 * 4 # 5 antenna fields (CS001HBA0, CS001HBA1, CS002HBA0, CS002HBA1, RS205HBA) * 4 stokes

        subtask = self.create_subtask(specifications_doc)
        parset = convert_to_parset_dict(subtask)
        logger.info("test_flyseye parset:",parset)

        self.assertEqual(True,     parset["Cobalt.BeamFormer.flysEye"])
        self.assertEqual(True,     parset["Observation.DataProducts.Output_CoherentStokes.enabled"])
        self.assertEqual(nr_files, len(parset["Observation.DataProducts.Output_CoherentStokes.filenames"]))

        # check whether parset is accepted by the ResourceEstimator
        estimator = ObservationResourceEstimator()
        estimations = estimator.verify_and_estimate(parset)
        self.assertEqual([],       estimations["errors"])

        # check whether the ResourceEstimator agrees with our spec
        self.assertEqual(nr_files, estimations["estimates"][0]["output_files"]["cs"][0]["properties"]["nr_of_cs_files"] * estimations["estimates"][0]["resource_count"])
        self.assertEqual(1,        estimations["estimates"][0]["output_files"]["cs"][0]["properties"]["nr_of_cs_parts"])
        self.assertEqual(4,        estimations["estimates"][0]["output_files"]["cs"][0]["properties"]["nr_of_cs_stokes"])

    def test_beamformer(self):
        specifications_doc = self.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = False
        specifications_doc['stations']['digital_pointings'] = [
          { "name": "target1",
            "subbands": list(range(8))
          }
        ]

        specifications_doc['COBALT']['beamformer']['tab_pipelines'] = [
            { "coherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },
              "incoherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },

              "SAPs": [
                { "name": "target1",
                  "tabs": [
                    {
                      "coherent": True,
                      "pointing": { "angle1": 1.0, "angle2": 2.0 }
                    },
                    {
                      "coherent": False
                    },
                  ]
                }
              ]
            }
        ]

        nr_cs_files = 1 * 4 # 1 TAB * 4 stokes
        nr_is_files = 1 * 4 # 1 TAB * 4 stokes

        subtask = self.create_subtask(specifications_doc)
        parset = convert_to_parset_dict(subtask)
        logger.info("test_beamformer parset:",parset)

        self.assertEqual(True,        parset["Observation.DataProducts.Output_CoherentStokes.enabled"])
        self.assertEqual(nr_cs_files, len(parset["Observation.DataProducts.Output_CoherentStokes.filenames"]))
        self.assertEqual(True,        parset["Observation.DataProducts.Output_IncoherentStokes.enabled"])
        self.assertEqual(nr_is_files, len(parset["Observation.DataProducts.Output_IncoherentStokes.filenames"]))

        # check whether parset is accepted by the ResourceEstimator
        estimator = ObservationResourceEstimator()
        estimations = estimator.verify_and_estimate(parset)
        self.assertEqual([],       estimations["errors"])

        # check whether the ResourceEstimator agrees with our spec
        self.assertEqual(nr_cs_files, estimations["estimates"][0]["output_files"]["cs"][0]["properties"]["nr_of_cs_files"] * estimations["estimates"][0]["resource_count"])
        self.assertEqual(1,           estimations["estimates"][0]["output_files"]["cs"][0]["properties"]["nr_of_cs_parts"])
        self.assertEqual(4,           estimations["estimates"][0]["output_files"]["cs"][0]["properties"]["nr_of_cs_stokes"])

        self.assertEqual(nr_is_files, estimations["estimates"][1]["output_files"]["is"][0]["properties"]["nr_of_is_files"] * estimations["estimates"][1]["resource_count"])
        self.assertEqual(4,           estimations["estimates"][1]["output_files"]["is"][0]["properties"]["nr_of_is_stokes"])

class SIPadapterTest(unittest.TestCase):
    def test_simple_sip_generate_from_dataproduct(self):
        """
        Test if SIP is generated successfully when subtask, dataproduct and SAP objects are created
        Check some value in the SIP (xml) output
        Check that the SIP identifiers are in SIP (xml) output
        Check the number of SIP identifiers are increased with 3
        Check that all SIP identifiers are unique
        """
        subtask_template = models.SubtaskTemplate.objects.get(name='observation control')
        specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
        specifications_doc['stations']['filter'] = "HBA_210_250"
        feedback_template = models.DataproductFeedbackTemplate.objects.get(name='feedback')
        # feedback_doc = get_default_json_object_for_schema(feedback_template.schema)  # todo <- fix the default generator, for some reason it does not produce valid json here...
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [156], 'central_frequencies': [33593750.0], 'channel_width': 6103.515625, 'channels_per_subband': 32}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'angle3': 0, 'direction_type': 'J2000'}}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}, '$schema': 'http://127.0.0.1:8001/api/schemas/dataproductfeedbacktemplate/feedback/1#'}
        for dp in specifications_doc['stations']['digital_pointings']:
            dp['subbands'] = list(range(8))
        # Create SubTask(output)
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc)
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))
        # Create Dataproduct
        dataproduct: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(feedback_doc=feedback_doc, producer=subtask_output))

        # Create SAP
        sap_template = models.SAPTemplate.objects.get(name="SAP")
        specifications_doc = get_default_json_object_for_schema(sap_template.schema)
        sap = models.SAP.objects.create(specifications_doc=specifications_doc, specifications_template=sap_template)
        dataproduct.sap = sap
        sap.save()

        sip = generate_sip_for_dataproduct(dataproduct)

        # double-check that SIP contains values from feedback and specifications docs
        self.assertIn(str(feedback_doc['frequency']['channel_width']), sip.get_prettyxml())
        self.assertIn(str(feedback_doc['time']['start_time']), sip.get_prettyxml())
        self.assertIn(constants.FILTERSELECTIONTYPE_210_250_MHZ, sip.get_prettyxml()) # specifications_doc: "HBA_210_250"

        self.assertIn(str(subtask.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(dataproduct.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(sap.global_identifier.unique_identifier), sip.get_prettyxml())


class FeedbackAdapterTest(unittest.TestCase):

    feedback_pipe_complete = """LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].centralFrequency=33593750.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].channelWidth=6103.515625
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].channelsPerSubband=32
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].duration=5.02732992172
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].fileFormat=AIPS++/CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].filename=L99307_SB000_uv.dppp.MS
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].integrationInterval=2.00278016
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].location=locus001:/data/L99307
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].returncode=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].size=15606123742
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].startTime=2013-02-16T17:00:00.000
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].stationSubband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].storageWriter=CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].storageWriterVersion=2.2.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].subband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].centralFrequency=33789062.5
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelWidth=6103.515625
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelsPerSubband=32
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].duration=5.02513194084
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].fileFormat=AIPS++/CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].filename=L99307_SB001_uv.dppp.MS
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].integrationInterval=2.00278016
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].location=locus003:/data/L99307
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].returncode=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].size=15606156518
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].startTime=2013-02-16T17:00:00.000
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].stationSubband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriter=CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriterVersion=2.2.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].subband=1
Observation.DataProducts.nrOfOutput_Beamformed_=0
Observation.DataProducts.nrOfOutput_Correlated_=2
_isCobalt=T
feedback_version=03.01.00
"""

    feedback_pipe_incomplete = """LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].centralFrequency=33789062.5
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelWidth=6103.515625
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelsPerSubband=32
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].duration=5.02513194084
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].fileFormat=AIPS++/CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].filename=L99307_SB001_uv.dppp.MS
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].integrationInterval=2.00278016
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].location=locus003:/data/L99307
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].returncode=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].size=15606156518
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].startTime=2013-02-16T17:00:00.000
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].stationSubband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriter=CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriterVersion=2.2.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].subband=1
Observation.DataProducts.nrOfOutput_Beamformed_=0
Observation.DataProducts.nrOfOutput_Correlated_=2
_isCobalt=T
feedback_version=03.01.00"""

    feedback_obs_complete = """Observation.Correlator.channelWidth=3051.7578125
Observation.Correlator.channelsPerSubband=64
Observation.Correlator.integrationInterval=1.00663296
Observation.DataProducts.Output_Correlated_[0].SAP=0
Observation.DataProducts.Output_Correlated_[0].centralFrequency=30468750.000000
Observation.DataProducts.Output_Correlated_[0].channelWidth=3051.757812
Observation.DataProducts.Output_Correlated_[0].channelsPerSubband=64
Observation.DataProducts.Output_Correlated_[0].duration=0
Observation.DataProducts.Output_Correlated_[0].fileFormat=AIPS++/CASA
Observation.DataProducts.Output_Correlated_[0].filename=L220133_SAP000_SB000_uv.MS
Observation.DataProducts.Output_Correlated_[0].integrationInterval=1.006633
Observation.DataProducts.Output_Correlated_[0].location=CEP2:/data/L220133/
Observation.DataProducts.Output_Correlated_[0].percentageWritten=0
Observation.DataProducts.Output_Correlated_[0].size=0
Observation.DataProducts.Output_Correlated_[0].startTime=2014-04-18 15:02:00
Observation.DataProducts.Output_Correlated_[0].stationSubband=156
Observation.DataProducts.Output_Correlated_[0].storageWriter=LOFAR
Observation.DataProducts.Output_Correlated_[0].storageWriterVersion=3
Observation.DataProducts.Output_Correlated_[0].subband=0
Observation.DataProducts.Output_Correlated_[1].SAP=0
Observation.DataProducts.Output_Correlated_[1].centralFrequency=30664062.500000
Observation.DataProducts.Output_Correlated_[1].channelWidth=3051.757812
Observation.DataProducts.Output_Correlated_[1].channelsPerSubband=64
Observation.DataProducts.Output_Correlated_[1].duration=0
Observation.DataProducts.Output_Correlated_[1].fileFormat=AIPS++/CASA
Observation.DataProducts.Output_Correlated_[1].filename=L220133_SAP000_SB001_uv.MS
Observation.DataProducts.Output_Correlated_[1].integrationInterval=1.006633
Observation.DataProducts.Output_Correlated_[1].location=CEP2:/data/L220133/
Observation.DataProducts.Output_Correlated_[1].percentageWritten=0
Observation.DataProducts.Output_Correlated_[1].size=0
Observation.DataProducts.Output_Correlated_[1].startTime=2014-04-18 15:02:00
Observation.DataProducts.Output_Correlated_[1].stationSubband=157
Observation.DataProducts.Output_Correlated_[1].storageWriter=LOFAR
Observation.DataProducts.Output_Correlated_[1].storageWriterVersion=3
Observation.DataProducts.Output_Correlated_[1].subband=1
Observation.DataProducts.nrOfOutput_Beamformed_=0
Observation.DataProducts.nrOfOutput_Correlated_=2
_isCobalt=T
"""

    def test_append_to_subtask_raw_feedback_raises_on_wrong_subtask_state(self):
        subtask_data = Subtask_test_data()
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        with self.assertRaises(SubtaskInvalidStateException) as cm:
            append_to_subtask_raw_feedback(subtask, parameterset.fromString(self.feedback_obs_complete))

        self.assertIn("Cannot process feedback", str(cm.exception))


    def test_process_feedback_into_subtask_dataproducts_raises_on_wrong_subtask_state(self):
        subtask_data = Subtask_test_data()
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        with self.assertRaises(SubtaskInvalidStateException) as cm:
            process_feedback_into_subtask_dataproducts(subtask, parameterset.fromString(self.feedback_obs_complete))

        self.assertIn("Cannot process feedback", str(cm.exception))


    def test_process_feedback_for_subtask_and_set_to_finished_if_complete_raises_on_wrong_subtask_state(self):
        subtask_data = Subtask_test_data()
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        with self.assertRaises(SubtaskInvalidStateException) as cm:
            process_feedback_for_subtask_and_set_to_finished_if_complete(subtask, self.feedback_obs_complete)

        self.assertIn("Cannot process feedback", str(cm.exception))


    def test_generate_dataproduct_feedback_from_subtask_feedback_and_set_finished_fails_on_incomplete_feedback(self):
        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value='finishing'),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='observation control'))
        subtask_obs:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_obs_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_obs))

        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value='finishing'),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='pipeline control'))
        subtask_pipe: models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_pipe_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_pipe))

        test_dir = "/tmp/test/data/%s" % uuid.uuid4()
        empty_feedback_template = models.DataproductFeedbackTemplate.objects.get(name='empty')
        dataproduct_obs_out1:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB000_uv.MS', directory=test_dir, producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_obs_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB001_uv.MS', directory=test_dir, producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out1: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB000_uv.dppp.MS', directory=test_dir, producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB001_uv.dppp.MS', directory=test_dir, producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        models.DataproductTransform.objects.create(input=dataproduct_obs_out1, output=dataproduct_pipe_out1, identity=True)
        models.DataproductTransform.objects.create(input=dataproduct_obs_out2, output=dataproduct_pipe_out2, identity=True)

        # assert dataproducts have no feedback docs before conversion
        for dataproduct in [dataproduct_obs_out1, dataproduct_obs_out2, dataproduct_pipe_out1, dataproduct_pipe_out2]:
            self.assertNotIn('percentage_written', dataproduct.feedback_doc)

        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_obs, self.feedback_obs_complete)
        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_pipe, self.feedback_pipe_incomplete) # <--- Note: test for incomplete feedback

        # assert not in FINISHED state
        self.assertEqual(models.SubtaskState.objects.get(value='finishing'), subtask_pipe.state)
        self.assertFalse(subtask_pipe.is_feedback_complete)

        # assert raw_feedback stored
        self.assertEqual(self.feedback_pipe_incomplete.strip(), subtask_pipe.raw_feedback.strip())

    def test_generate_dataproduct_feedback_from_subtask_feedback_and_set_finished_after_reprocessing(self):
        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value='finishing'),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='observation control'))
        subtask_obs:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_obs_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_obs))

        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value='finishing'),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='pipeline control'))
        subtask_pipe: models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_pipe_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_pipe))

        test_dir = "/tmp/test/data/%s" % uuid.uuid4()
        empty_feedback_template = models.DataproductFeedbackTemplate.objects.get(name='empty')
        dataproduct_obs_out1:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB000_uv.MS', directory=test_dir, producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_obs_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB001_uv.MS', directory=test_dir, producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out1: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB000_uv.dppp.MS', directory=test_dir, producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB001_uv.dppp.MS', directory=test_dir, producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        models.DataproductTransform.objects.create(input=dataproduct_obs_out1, output=dataproduct_pipe_out1, identity=True)
        models.DataproductTransform.objects.create(input=dataproduct_obs_out2, output=dataproduct_pipe_out2, identity=True)

        # assert dataproducts have no feedback docs before conversion
        for dataproduct in [dataproduct_obs_out1, dataproduct_obs_out2, dataproduct_pipe_out1, dataproduct_pipe_out2]:
            self.assertNotIn('percentage_written', dataproduct.feedback_doc)

        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_obs, self.feedback_obs_complete)
        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_pipe, self.feedback_pipe_incomplete) # <--- Note: test for incomplete feedback

        # assert not in FINISHED state
        self.assertEqual(models.SubtaskState.objects.get(value='finishing'), subtask_pipe.state)
        self.assertFalse(subtask_pipe.is_feedback_complete)

        # assert raw_feedback stored
        self.assertEqual(self.feedback_pipe_incomplete.strip(), subtask_pipe.raw_feedback.strip())

        # update the complete raw_feedback (for example by a support engineer)
        subtask_pipe.raw_feedback = self.feedback_pipe_complete
        subtask_pipe.save()

        reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete(subtask_pipe)

        # Now it should be finished and complete
        self.assertEqual(models.SubtaskState.objects.get(value='finished'), subtask_pipe.state)
        self.assertTrue(subtask_pipe.is_feedback_complete)

    def test_generate_dataproduct_feedback_from_subtask_feedback_and_set_finished(self):
        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value='finishing'),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='observation control'))
        subtask_obs:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_obs_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_obs))

        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value='finishing'),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='pipeline control'))
        subtask_pipe: models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_pipe_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_pipe))

        empty_feedback_template = models.DataproductFeedbackTemplate.objects.get(name='empty')
        dataproduct_obs_out1:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB000_uv.MS', producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_obs_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB001_uv.MS', producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out1: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB000_uv.dppp.MS', producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB001_uv.dppp.MS', producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        models.DataproductTransform.objects.create(input=dataproduct_obs_out1, output=dataproduct_pipe_out1, identity=True)
        models.DataproductTransform.objects.create(input=dataproduct_obs_out2, output=dataproduct_pipe_out2, identity=True)

        # assert dataproducts have no feedback docs before conversion
        for dataproduct in [dataproduct_obs_out1, dataproduct_obs_out2, dataproduct_pipe_out1, dataproduct_pipe_out2]:
            self.assertNotIn('percentage_written', dataproduct.feedback_doc)

        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_obs, self.feedback_obs_complete)
        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_pipe, self.feedback_pipe_complete)

        # reload dataproducts and assert dataproduct feedback docs have feedback after conversion
        for dataproduct in [dataproduct_obs_out1, dataproduct_obs_out2, dataproduct_pipe_out1, dataproduct_pipe_out2]:
            dataproduct.refresh_from_db()
            self.assertIsNotNone(dataproduct.feedback_doc)
            self.assertIn('percentage_written', dataproduct.feedback_doc)

        # assert correct relations of feedback docs
        self.assertNotEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                            dataproduct_obs_out2.feedback_doc['frequency']['subbands'])
        self.assertNotEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                            dataproduct_pipe_out2.feedback_doc['frequency']['subbands'])
        self.assertEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                         dataproduct_pipe_out1.feedback_doc['frequency']['subbands'])
        self.assertEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                         dataproduct_pipe_out1.feedback_doc['frequency']['subbands'])

        # assert FINISHED states
        self.assertEqual(models.SubtaskState.objects.get(value='finished'), subtask_obs.state)
        self.assertEqual(models.SubtaskState.objects.get(value='finished'), subtask_pipe.state)

        self.assertTrue(subtask_obs.is_feedback_complete)
        self.assertTrue(subtask_pipe.is_feedback_complete)

        # assert raw_feedback stored
        self.assertEqual(self.feedback_obs_complete.strip(), subtask_obs.raw_feedback.strip())
        self.assertEqual(self.feedback_pipe_complete.strip(), subtask_pipe.raw_feedback.strip())



if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
