#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $


# This functional test talks to the API like a regular user would.
# It is supposed to cover all REST http methods for all ViewSets.
# I am still a bit under the impression that we re-test Django functionality that we can expect to just work
# with some of these tests. On the other hand a lot of these provide us a nice basis for differentiating out
# behavior in a controlled way.
# We should probably also fully test behavior wrt mandatory and nullable fields.

from datetime import datetime
import unittest
from unittest import mock
import logging
import requests

logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator

from django.test import TestCase

from django.contrib.auth.models import User, Group, Permission

class ProjectPermissionTestCase(TestCase):
    # This tests that the project permissions are enforced in light of the project roles that are externally provided
    # for the user through the user admin. This test does not rely on the project permissions as defined in the system,
    # but project permissions are created specifically for this test.


    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        # todo: this mock does not work. Fix and remove the mock response for user paulus from the actual permissions module
        # mock the project roles usually provided by the identity management system
        cls.project_permission_patcher = mock.patch('lofar.sas.tmss.tmss.tmssapp.viewsets.get_project_roles_for_user')  # todo: fix namespace so we get the get_project_roles that gets actually used
        cls.project_permission_mock = cls.project_permission_patcher.start()
        cls.project_permission_mock.return_value = ({'project': 'test_user_is_shared_support', 'role': 'shared_support_user'},
                                                    {'project': 'test_user_is_contact', 'role': 'contact_author'})

        # create some stuff as the standard super user, as setup for the tests below
        cls.test_data_creator = TMSSRESTTestDataCreator(BASE_URL, AUTH)
        cycle_url = cls.test_data_creator.cached_cycle_url

        # create projects with magic names for which permission exists (or which have no whitelisted generic name)
        cls.project_shared_support_user_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.Project(name='test_user_is_shared_support', cycle_urls=[cycle_url]), '/project/')
        cls.project_contact_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.Project(name='test_user_is_contact', cycle_urls=[cycle_url]), '/project/')
        cls.project_forbidden_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.Project(name='forbidden', cycle_urls=[cycle_url]), '/project/')


        # user is shared_support_user
        cls.scheduling_set_shared_support_user_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.SchedulingSet(project_url=cls.project_shared_support_user_url), '/scheduling_set/')
        cls.scheduling_unit_draft_shared_support_user_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.SchedulingUnitDraft(scheduling_set_url=cls.scheduling_set_shared_support_user_url), '/scheduling_unit_draft/')

        # user is contact
        cls.scheduling_set_contact_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.SchedulingSet(project_url=cls.project_contact_url), '/scheduling_set/')
        cls.scheduling_unit_draft_contact_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.SchedulingUnitDraft(scheduling_set_url=cls.scheduling_set_contact_url), '/scheduling_unit_draft/')

        # user has no role
        cls.scheduling_set_forbidden_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.SchedulingSet(project_url=cls.project_forbidden_url), '/scheduling_set/')
        cls.scheduling_unit_draft_forbidden_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.SchedulingUnitDraft(scheduling_set_url=cls.scheduling_set_contact_url), '/scheduling_unit_draft/')

        # create the required permission entries to control what endpoint action requires which project role
        shared_support_user_role_url = BASE_URL + '/project_role/shared_support_user/'
        cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.ProjectPermission(name='taskdraft', GET=[shared_support_user_role_url], POST=[shared_support_user_role_url]), '/project_permission/')
        cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.ProjectPermission(name='taskdraft-create_task_blueprint', GET=[shared_support_user_role_url]), '/project_permission/')

        cls.task_template_url = cls.test_data_creator.post_data_and_get_url(cls.test_data_creator.TaskTemplate(), '/task_template/')

        # create a new test_data_creator with the regular 'paulus' user, not superuser as in other tests, so that project permissions of tht user are checked.
        cls.auth = requests.auth.HTTPBasicAuth('paulus', 'pauluspass')
        cls.test_data_creator = TMSSRESTTestDataCreator(BASE_URL, cls.auth)


    # TaskDraft

    def test_task_draft_GET_works_if_user_has_permission_for_related_project(self):
        # create task draft connected to project where we have 'shared_support_user' role
        taskdraft_test_data = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_shared_support_user_url, template_url=self.task_template_url)
        taskdraft_url = POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data, 201, taskdraft_test_data)['url']

        # make sure we can access it
        GET_and_assert_equal_expected_code(self, taskdraft_url, 200, auth=self.auth)

    def test_task_draft_GET_raises_error_if_user_has_no_permission_for_related_project(self):
        # create task draft connected to project where we have no role
        taskdraft_test_data = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_forbidden_url, template_url=self.task_template_url)
        taskdraft_url = POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data, 201, taskdraft_test_data)['url']

        # make sure we cannot access it
        GET_and_assert_equal_expected_code(self, taskdraft_url, 403, auth=self.auth)

    def test_task_draft_GET_raises_error_if_user_has_permission_for_related_project_but_with_wrong_role(self):
        # create task draft connected to project where we have Contact Author role  (Task Draft access requires role 'shared_support_user')
        taskdraft_test_data = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_contact_url, template_url=self.task_template_url)
        taskdraft_url = POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data, 201, taskdraft_test_data)['url']

        # make sure we cannot access it
        GET_and_assert_equal_expected_code(self, taskdraft_url, 403, auth=self.auth)

    def test_GET_task_draft_list_returns_filtered_list_reflecting_user_permission_for_related_projects(self):
        nbr_results = GET_and_assert_equal_expected_code(self, BASE_URL + '/task_draft/', 200, auth=self.auth)["count"] # note: this does not guarantee the correct number with permission-based filtering: nbr_results = models.TaskDraft.objects.count()

        # create task draft connected to project where we have sufficient role
        taskdraft_test_data_shared_support_user = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_shared_support_user_url, template_url=self.task_template_url)
        POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data_shared_support_user, 201, taskdraft_test_data_shared_support_user)

        # create task draft connected to project where we have unsufficient contact role
        taskdraft_test_data_contact = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_contact_url, template_url=self.task_template_url)
        POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data_contact, 201, taskdraft_test_data_contact)

        # create task draft connected to project where we have no role
        taskdraft_test_data_forbidden = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_contact_url, template_url=self.task_template_url)
        POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data_forbidden, 201, taskdraft_test_data_forbidden)

        # make sure the list contains only the one more item we have permission for
        GET_and_assert_in_expected_response_result_list(self, BASE_URL + '/task_draft/', taskdraft_test_data_shared_support_user, nbr_results + 1, auth=self.auth)

    def test_task_draft_POST_raises_error_if_user_has_no_permission_for_related_project(self):
        # try to create a task draft connected to project where we have no role and make sure that fails
        taskdraft_test_data = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_forbidden_url, template_url=self.task_template_url)
        POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data, 403, taskdraft_test_data, auth=self.auth)

    def test_task_draft_POST_works_if_user_has_permission_for_related_project(self):
        # create task draft connected to project where we have 'shared_support_user' role
        taskdraft_test_data = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_shared_support_user_url, template_url=self.task_template_url)
        POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data, 201, taskdraft_test_data, auth=self.auth)

    # TaskDraft.actions

    def test_task_draft_create_task_blueprint_GET_raises_error_if_user_has_no_permission_for_related_project(self):
        # create task draft connected to project where we have no role
        taskdraft_test_data = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_forbidden_url, template_url=self.task_template_url)
        taskdraft_url = POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data, 201, taskdraft_test_data)['url']

        # make sure we cannot create a blueprint from it
        GET_and_assert_equal_expected_code(self, taskdraft_url + '/create_task_blueprint/', 403, auth=self.auth)

    def test_task_draft_create_task_blueprint_GET_works_if_user_has_permission_for_related_project(self):
        # create task draft connected to project where we have 'shared_support_user' role
        taskdraft_test_data = self.test_data_creator.TaskDraft(scheduling_unit_draft_url=self.scheduling_unit_draft_shared_support_user_url, template_url=self.task_template_url)
        taskdraft_url = POST_and_assert_expected_response(self, BASE_URL + '/task_draft/', taskdraft_test_data, 201, taskdraft_test_data)['url']

        # make sure we cannot create a blueprint from it
        GET_and_assert_equal_expected_code(self, taskdraft_url + '/create_task_blueprint/', 201, auth=self.auth)

    # todo: add tests for other models with project permissions


if __name__ == "__main__":
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=logging.INFO)
    unittest.main()

