#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
from unittest import mock

import logging
logger = logging.getLogger('lofar'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# before we import any django modules the DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS need to be known/set.
# import and start an isolated RATestEnvironment and TMSSTestEnvironment (with fresh database and attached django and ldap server on free ports)
# this automagically sets the required  DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS envvars.
from lofar.sas.tmss.test.test_utils import TMSSTestEnvironment

tmss_test_env = TMSSTestEnvironment(populate_schemas=True, populate_test_data=False, start_ra_test_environment=True,
                                    start_postgres_listener=False, start_subtask_scheduler=False, start_dynamic_scheduler=False,
                                    enable_viewflow=False)

try:
    tmss_test_env.start()
except Exception as e:
    logger.exception(e)

    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()


from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
test_data_creator = TMSSRESTTestDataCreator(tmss_test_env.django_server.url, (tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password))
from datetime import datetime, timedelta
from lofar.sas.resourceassignment.resourceassigner.rarpc import RARPC
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.subtasks import *
from lofar.sas.tmss.tmss.tmssapp.tasks import *


def create_subtask_object_for_testing(subtask_type_value, subtask_state_value):
    """
    Helper function to create a subtask object for testing with given subtask value and subtask state value
    as string (no object)
    For these testcases 'pipeline control' and 'observation control' is relevant
    """
    task_blueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(specifications_template=models.TaskTemplate.objects.get(name='target observation' if subtask_type_value=='observation' else 'preprocessing pipeline')))
    subtask_template_obj = models.SubtaskTemplate.objects.get(name="%s control" % subtask_type_value)
    subtask_state_obj = models.SubtaskState.objects.get(value=subtask_state_value)
    subtask_data = Subtask_test_data(subtask_template=subtask_template_obj, state=subtask_state_obj, task_blueprint=task_blueprint)
    return models.Subtask.objects.create(**subtask_data)


def create_reserved_stations_for_testing(station_list):
    """
    Helper function to create stations in reservation, in other words assigned in Resource Assigner
    :param station_list: List of station names to assign
    """
    with RARPC.create() as rarpc:
        ra_spec = {'task_type': 'reservation',
                   'task_subtype': 'maintenance',
                   'status': 'prescheduled',
                   'starttime': datetime.utcnow() - timedelta(hours=1),
                   'endtime': datetime.utcnow() + timedelta(hours=2),
                   'cluster': None,
                   'specification': {}}
        inner_spec = {'Observation.VirtualInstrument.stationList': station_list,
                      'Observation.startTime': ra_spec['starttime'],
                      'Observation.endTime': ra_spec['starttime']}
        ra_spec['specification'] = inner_spec
        assigned = rarpc.do_assignment(ra_spec)
        return assigned


class SchedulingTest(unittest.TestCase):
    def setUp(self):
        # clean all specs/tasks/claims in RADB (cascading delete)
        for spec in tmss_test_env.ra_test_environment.radb.getSpecifications():
            tmss_test_env.ra_test_environment.radb.deleteSpecification(spec['id'])

        DataproductTransform.objects.all().delete()
        Dataproduct.objects.all().delete()
        SubtaskInput.objects.all().delete()
        SubtaskOutput.objects.all().delete()
        Subtask.objects.all().delete()

        test_data_creator.wipe_cache()


    def test_schedule_observation_subtask_with_enough_resources_available(self):
        with tmss_test_env.create_tmss_client() as client:
            task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(task_blueprint_data, '/task_blueprint/')
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url=cluster_url,
                                                     start_time=datetime.utcnow()+timedelta(minutes=5),
                                                     task_blueprint_url=task_blueprint['url'])
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url']), '/subtask_output/')

            client.set_subtask_status(subtask_id, 'defined')
            subtask = client.schedule_subtask(subtask_id)

            self.assertEqual('scheduled', subtask['state_value'])
            self.assertEqual('scheduled', tmss_test_env.ra_test_environment.radb.getTask(tmss_id=subtask_id)['status'])

    def test_schedule_observation_subtask_with_one_blocking_reservation_failed(self):
        """
        Set (Resource Assigner) station CS001 to reserved
        Schedule subtask with station CS001
        Check if schedule of the subtask fail
        """
        self.assertTrue(create_reserved_stations_for_testing(['CS001']))

        with tmss_test_env.create_tmss_client() as client:
            task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(task_blueprint_data, '/task_blueprint/')
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url=cluster_url,
                                                     start_time=datetime.utcnow() + timedelta(minutes=5),
                                                     task_blueprint_url=task_blueprint['url'])
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url']), '/subtask_output/')

            client.set_subtask_status(subtask_id, 'defined')

            with self.assertRaises(Exception):
                client.schedule_subtask(subtask_id)

            subtask = client.get_subtask(subtask_id)
            self.assertEqual('error', subtask['state_value'])
            self.assertEqual('conflict', tmss_test_env.ra_test_environment.radb.getTask(tmss_id=subtask_id)['status'])

    def test_schedule_observation_subtask_with_blocking_reservations_failed(self):
        """
        Set (Resource Assigner) station CS001, CS002, CS401, CS501 to reserved
        Schedule subtask with stations CS001, CS002, CS401
        Check if schedule of the subtask fail
        """
        self.assertTrue(create_reserved_stations_for_testing(['CS001','CS002','CS501','CS401' ]))

        with tmss_test_env.create_tmss_client() as client:
            task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(task_blueprint_data, '/task_blueprint/')

            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            spec['stations']['station_list'] = ['CS001', 'CS002', 'CS401']

            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url=cluster_url,
                                                     start_time=datetime.utcnow() + timedelta(minutes=5),
                                                     task_blueprint_url=task_blueprint['url'])
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url']), '/subtask_output/')

            client.set_subtask_status(subtask_id, 'defined')

            with self.assertRaises(Exception):
                client.schedule_subtask(subtask_id)

            subtask = client.get_subtask(subtask_id)
            self.assertEqual('error', subtask['state_value'])
            ra_task = tmss_test_env.ra_test_environment.radb.getTask(tmss_id=subtask_id)
            self.assertIsNotNone(ra_task)
            self.assertEqual('conflict', ra_task['status'])

    def test_schedule_observation_subtask_with_blocking_reservation_ok(self):
        """
        Set (Resource Assigner) station CS001 to reserved
        Schedule subtask with station CS001, CS002, CS003
        Check if schedule of the subtasks do not fail (it can schedule with station CS002 and CS003)
        """
        self.assertTrue(create_reserved_stations_for_testing(['CS001','CS003']))

        with tmss_test_env.create_tmss_client() as client:
            task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(task_blueprint_data,'/task_blueprint/')
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']
            spec['stations']['station_list'] = ['CS001', 'CS002', 'CS003']
            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url=cluster_url,
                                                     start_time=datetime.utcnow()+timedelta(minutes=5),
                                                     task_blueprint_url=task_blueprint['url'])
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url']),
                                                    '/subtask_output/')

            client.set_subtask_status(subtask_id, 'defined')

            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])
            self.assertEqual('scheduled', tmss_test_env.ra_test_environment.radb.getTask(tmss_id=subtask_id)['status'])

    def test_schedule_pipeline_subtask_with_enough_resources_available(self):
        with tmss_test_env.create_tmss_client() as client:
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            # setup: first create an observation, so the pipeline can have input.
            obs_task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            obs_task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(obs_task_blueprint_data, '/task_blueprint/')
            obs_subtask_template = client.get_subtask_template("observation control")
            obs_spec = get_default_json_object_for_schema(obs_subtask_template['schema'])
            obs_spec['stations']['digital_pointings'][0]['subbands'] = [0]

            obs_subtask_data = test_data_creator.Subtask(specifications_template_url=obs_subtask_template['url'],
                                                         specifications_doc=obs_spec,
                                                         cluster_url=cluster_url,
                                                         task_blueprint_url=obs_task_blueprint['url'])
            obs_subtask = test_data_creator.post_data_and_get_response_as_json_object(obs_subtask_data, '/subtask/')
            obs_subtask_output_url = test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=obs_subtask['url']), '/subtask_output/')
            test_data_creator.post_data_and_get_url(test_data_creator.Dataproduct(filename="L%s_SB000.MS"%obs_subtask['id'],
                                                                                  subtask_output_url=obs_subtask_output_url), '/dataproduct/')

            # now create the pipeline...
            pipe_task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="preprocessing pipeline")['url'])
            pipe_task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(pipe_task_blueprint_data, '/task_blueprint/')

            pipe_subtask_template = client.get_subtask_template("pipeline control")
            pipe_spec = get_default_json_object_for_schema(pipe_subtask_template['schema'])

            pipe_subtask_data = test_data_creator.Subtask(specifications_template_url=pipe_subtask_template['url'],
                                                          specifications_doc=pipe_spec,
                                                          task_blueprint_url=pipe_task_blueprint['url'],
                                                          cluster_url=cluster_url)
            pipe_subtask = test_data_creator.post_data_and_get_response_as_json_object(pipe_subtask_data, '/subtask/')

            # ...and connect it to the observation
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskInput(subtask_url=pipe_subtask['url'], subtask_output_url=obs_subtask_output_url), '/subtask_input/')
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=pipe_subtask['url']), '/subtask_output/')

            for predecessor in client.get_subtask_predecessors(pipe_subtask['id']):
                client.set_subtask_status(predecessor['id'], 'finished')

            client.set_subtask_status(pipe_subtask['id'], 'defined')
            subtask = client.schedule_subtask(pipe_subtask['id'])

            self.assertEqual('scheduled', subtask['state_value'])
            self.assertEqual('scheduled', tmss_test_env.ra_test_environment.radb.getTask(tmss_id=pipe_subtask['id'])['status'])

    def test_schedule_ingest_subtask(self):
        with tmss_test_env.create_tmss_client() as client:
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            # setup: first create an observation, so the ingest can have input.
            obs_subtask_template = client.get_subtask_template("observation control")
            obs_spec = get_default_json_object_for_schema(obs_subtask_template['schema'])
            obs_spec['stations']['digital_pointings'][0]['subbands'] = [0]

            obs_subtask_data = test_data_creator.Subtask(specifications_template_url=obs_subtask_template['url'],
                                                         specifications_doc=obs_spec,
                                                         cluster_url=cluster_url,
                                                         task_blueprint_url=test_data_creator.post_data_and_get_url(test_data_creator.TaskBlueprint(), '/task_blueprint/'))
            obs_subtask = test_data_creator.post_data_and_get_response_as_json_object(obs_subtask_data, '/subtask/')
            obs_subtask_output_url = test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=obs_subtask['url']), '/subtask_output/')
            test_data_creator.post_data_and_get_url(test_data_creator.Dataproduct(filename="L%s_SB000.MS"%obs_subtask['id'],
                                                    subtask_output_url=obs_subtask_output_url), '/dataproduct/')

            # now create the ingest...
            ingest_subtask_template = client.get_subtask_template("ingest control")
            ingest_spec = get_default_json_object_for_schema(ingest_subtask_template['schema'])

            ingest_subtask_data = test_data_creator.Subtask(specifications_template_url=ingest_subtask_template['url'],
                                                          specifications_doc=ingest_spec,
                                                          task_blueprint_url=obs_subtask['task_blueprint'],
                                                          cluster_url=cluster_url)
            ingest_subtask = test_data_creator.post_data_and_get_response_as_json_object(ingest_subtask_data, '/subtask/')

            # ...and connect it to the observation
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskInput(subtask_url=ingest_subtask['url'], subtask_output_url=obs_subtask_output_url), '/subtask_input/')
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=ingest_subtask['url']), '/subtask_output/')

            for predecessor in client.get_subtask_predecessors(ingest_subtask['id']):
                client.set_subtask_status(predecessor['id'], 'finished')
            client.set_subtask_status(ingest_subtask['id'], 'defined')

            task_blueprint = client.get_url_as_json_object(ingest_subtask['task_blueprint'])
            schedulingunit_blueprint = client.get_url_as_json_object(task_blueprint['scheduling_unit_blueprint'])

            # first, make sure we need but do not have ingest persmission...
            client.session.patch(schedulingunit_blueprint['url'], json={'ingest_permission_required': True, 'ingest_permission_granted_since': None})

            with self.assertRaises(Exception) as context:
                subtask = client.schedule_subtask(ingest_subtask['id'])
                self.assertTrue('permission' in str(context.exception))

            subtask = client.get_subtask(ingest_subtask['id'])
            self.assertEqual('defined', subtask['state_value'])

            # now grant permission...
            client.session.patch(schedulingunit_blueprint['url'], json={'ingest_permission_required': True, 'ingest_permission_granted_since': datetime.utcnow().isoformat()})

            subtask = client.schedule_subtask(ingest_subtask['id'])

            self.assertEqual('scheduled', subtask['state_value'])
            self.assertEqual(models.Subtask.objects.get(id=ingest_subtask['id']).inputs.first().dataproducts.count(), 1)


    def test_schedule_schedulingunit_enough_resources_available(self):
        '''similar test as test_schedule_pipeline_subtask_with_enough_resources_available, but now created from a scheduling_unit'''
        with tmss_test_env.create_tmss_client() as client:
            scheduling_unit_template = client.get_schedulingunit_template("scheduling unit")
            scheduling_unit_doc = get_default_json_object_for_schema(scheduling_unit_template['schema'])

            # define an observation without QA
            obs_task = get_default_json_object_for_schema(client.get_task_template(name="target observation")['schema'])
            obs_task['QA']['plots']['enabled'] = False
            obs_task['QA']['file_conversion']['enabled'] = False
            obs_task['SAPs'] = [{ 'subbands': [0,1] }]
            scheduling_unit_doc['tasks']["Observation"] = {"specifications_doc": obs_task,
                                                           "specifications_template": "target observation"}

            # define a pipeline
            scheduling_unit_doc['tasks']["Pipeline"] = { "specifications_doc": get_default_json_object_for_schema(client.get_task_template(name="preprocessing pipeline")['schema']),
                                                         "specifications_template": "preprocessing pipeline"}

            # connect obs to pipeline
            scheduling_unit_doc['task_relations'].append({"producer": "Observation",
                                                          "consumer": "Pipeline",
                                                          "input": { "role": "input", "datatype": "visibilities" },
                                                          "output": { "role": "correlator", "datatype": "visibilities" },
                                                          "dataformat": "MeasurementSet",
                                                          "selection_doc": {},
                                                          "selection_template": "all" })

            # submit
            scheduling_unit_draft_data = test_data_creator.SchedulingUnitDraft(template_url=scheduling_unit_template['url'],
                                                                               requirements_doc=scheduling_unit_doc)
            scheduling_unit_draft = test_data_creator.post_data_and_get_response_as_json_object(scheduling_unit_draft_data, '/scheduling_unit_draft/')

            # create the whole blueprints tree...
            scheduling_unit_blueprint = client.create_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft['id'])

            # fetch the created task_blueprints
            task_blueprints = [client.get_url_as_json_object(task_blueprint_url) for task_blueprint_url in scheduling_unit_blueprint['task_blueprints']]
            self.assertEqual(2, len(task_blueprints))
            # and make sure they are ordered correctly
            if "Pipeline" in task_blueprints[0]['name']:
                task_blueprints.reverse()

            for task_blueprint in task_blueprints:
                self.assertEqual(1, len(task_blueprint['subtasks']))
                subtask = client.get_url_as_json_object(task_blueprint['subtasks'][0])

                client.session.patch(subtask['url'], {'start_time': datetime.utcnow() + timedelta(minutes=5)})
                client.set_subtask_status(subtask['id'], 'defined')

                subtask = client.schedule_subtask(subtask['id'])

                self.assertEqual('scheduled', subtask['state_value'])
                self.assertEqual('scheduled', tmss_test_env.ra_test_environment.radb.getTask(tmss_id=subtask['id'])['status'])

                client.set_subtask_status(subtask['id'], 'finished')


class SubtaskInputOutputTest(unittest.TestCase):
    """
    Subtask Input and Output test
    These testcases are located in the t_scheduling module, because during scheduling the output
    dataproducts are assigned
    """

    def setUp(self) -> None:
        # make sure we're allowed to schedule
        setting = Setting.objects.get(name='dynamic_scheduling_enabled')
        setting.value = True
        setting.save()

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.assign_or_unassign_resources")
    def test_schedule_pipeline_subtask_filters_predecessor_output_dataproducts_for_input(self, assign_resources_mock):
        # setup:
        #   create observation subtask and outputs and dataproducts
        obs_st = create_subtask_object_for_testing('observation', 'finished')
        obs_out1 = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))
        obs_out2 = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))

        #   create connected pipeline subtask and inputs, specify input filtering
        pipe_st = create_subtask_object_for_testing('pipeline', 'defined')
        pipe_out = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=pipe_st)) # required by scheduling function
        pipe_in1 = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out1, selection_doc={'sap': ['target0']}))
        pipe_in2 = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out2, selection_doc={'sap': ['target1']}))

        #   create obs output dataproducts with specs we can filter on
        dp1_1 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_doc={'sap': ['target0']}))
        dp1_2 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_doc={'sap': ['target1']}))
        dp1_3 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_doc={'sap': ['target0']}))

        dp2_1 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out2, specifications_doc={'sap': ['target0']}))
        dp2_2 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out2, specifications_doc={'sap': ['target1']}))

        # trigger:
        #   schedule pipeline, which should attach the correct subset of dataproducts to the pipeline inputs
        schedule_pipeline_subtask(pipe_st)

        # assert:
        #   check correct input filtering
        self.assertEqual(set(pipe_in1.dataproducts.all()), {dp1_1, dp1_3})
        self.assertEqual(set(pipe_in2.dataproducts.all()), {dp2_2})


class SAPTest(unittest.TestCase):
    """
    SAP test
    These testcases are located in the t_scheduling module, because the SAP entries are created/assigned during scheduling
    """

    def setUp(self) -> None:
        # make sure we're allowed to schedule
        setting = Setting.objects.get(name='dynamic_scheduling_enabled')
        setting.value = True
        setting.save()

    def test_schedule_observation_subtask_creates_sap_with_correct_pointing(self):
        with tmss_test_env.create_tmss_client() as client:
            task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(task_blueprint_data, '/task_blueprint/')
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']
            pointing = {"angle1": 7.6, "angle2": 5.4, "direction_type": "J2000"}
            spec['stations']['digital_pointings'][0]['pointing'] = pointing

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url = cluster_url,
                                                     task_blueprint_url=task_blueprint['url'],
                                                     start_time=datetime.utcnow() + timedelta(minutes=5),
                                                     stop_time=datetime.utcnow() + timedelta(minutes=15))
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url']),
                                                    '/subtask_output/')

            subtask_model = models.Subtask.objects.get(id=subtask_id)
            self.assertEqual(0, subtask_model.output_dataproducts.values('sap').count())

            client.set_subtask_status(subtask_id, 'defined')
            subtask = client.schedule_subtask(subtask_id)

            self.assertEqual(1, subtask_model.output_dataproducts.values('sap').count())
            self.assertEqual(subtask_model.output_dataproducts.first().sap.specifications_doc['pointing']['angle1'], pointing['angle1'])
            self.assertEqual(subtask_model.output_dataproducts.first().sap.specifications_doc['pointing']['angle2'], pointing['angle2'])

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.assign_or_unassign_resources")
    def test_schedule_pipeline_subtask_copies_sap_from_input_to_output(self, assign_resources_mock):
        # setup:
        #   create observation subtask and outputs and dataproducts
        obs_st = create_subtask_object_for_testing('observation', 'finished')
        obs_out = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))

        #   create connected pipeline subtask and inputs, specify input filtering
        pipe_st = create_subtask_object_for_testing('pipeline', 'defined')
        pipe_out = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=pipe_st)) # required by scheduling function
        pipe_in = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out))

        #   create obs output dataproducts
        dp1_in = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out))
        dp2_in = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out))

        #   schedule pipeline, which should copy the SAP
        schedule_pipeline_subtask(pipe_st)

        # determine the newly created pipeline dataproducts
        dp1_out = DataproductTransform.objects.get(input=dp1_in).output
        dp2_out = DataproductTransform.objects.get(input=dp2_in).output

        # assert:
        self.assertEqual(dp1_in.sap, dp1_out.sap)
        self.assertEqual(dp2_in.sap, dp2_out.sap)


class TestWithUC1Specifications(unittest.TestCase):
    """
    The Setup will create Scheduling Unit Draft with UC1 strategy template
    It will use the function 'create_task_blueprints_and_subtasks_from_scheduling_unit_draft' which is then
    implicit tested.
    Create Task Blueprints and Subtasks:
        Observation Task 'Calibration 1'
            SubTask Observation Control
            SubTask QA File
            SubTask QA Plots
        Pipeline Task 'Pipeline 1'
            SubTask Pipeline Control
        Observation Task 'Target Observation'
            SubTask Observation Control
            SubTask QA File
            SubTask QA Plots
        Pipeline Task 'Pipeline target1'
            SubTask Pipeline Control
        Pipeline Task 'Pipeline target2'
            SubTask Pipeline Control
        Observation Task 'Calibration 2'
            SubTask Observation Control
            SubTask QA File
            SubTask QA Plots
        Pipeline Task 'Pipeline 2'
            SubTask Pipeline Control

    Note that this test requires Resource Assigner testenvironment being alive
    """
    @classmethod
    def setUpClass(cls) -> None:
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   requirements_doc=strategy_template.template,
                                   requirements_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   copy_reason=models.CopyReason.objects.get(value='template'),
                                   generator_instance_doc="para",
                                   copies=None,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft.refresh_from_db()
        cls.task_drafts = scheduling_unit_draft.task_drafts.all()
        cls.scheduling_unit_blueprints = scheduling_unit_draft.scheduling_unit_blueprints.all()
        cls.scheduling_unit_blueprint = cls.scheduling_unit_blueprints[0]
        cls.task_blueprints = cls.scheduling_unit_blueprint.task_blueprints.all()
        # SubtaskId of the first observation subtask
        observation_tbp = list(tb for tb in list(cls.task_blueprints) if tb.specifications_template.type.value == TaskType.Choices.OBSERVATION.value)
        observation_tbp.sort(key=lambda tb: tb.relative_start_time)
        cls.subtask_id_of_first_observation = list(st for st in observation_tbp[0].subtasks.all()
                                                   if st.specifications_template.type.value == SubtaskType.Choices.OBSERVATION.value)[0].id

    def setUp(self):
        # clean all specs/tasks/claims in RADB (cascading delete)
        for spec in tmss_test_env.ra_test_environment.radb.getSpecifications():
            tmss_test_env.ra_test_environment.radb.deleteSpecification(spec['id'])
        # Unschedule subtask, setting it back to 'defined', removing all dataproducts.
        for tb in self.task_blueprints:
            for subtask in tb.subtasks.all():
                if subtask.state.value == SubtaskState.Choices.SCHEDULED.value:
                    unschedule_subtask(subtask)
                if subtask.state.value == SubtaskState.Choices.ERROR.value:
                    subtask.state = SubtaskState.objects.get(value=SubtaskState.Choices.DEFINED.value)

                for output in subtask.outputs.all():
                    # delete all transforms (the producers of the output dataproducts), and the the dataproducts themselves
                    output.dataproducts.all().select_related('producers').delete()
                    output.dataproducts.all().delete()

                # start_time to now (and no stoptime)
                subtask.stop_time = None
                subtask.start_time = datetime.utcnow()
                subtask.save()



    def _schedule_subtask_with_failure(self, station_reserved):
        with tmss_test_env.create_tmss_client() as client:
            with self.assertRaises(Exception) as context:
                client.schedule_subtask(self.subtask_id_of_first_observation)
                self.assertTrue("There are more stations in conflict than the specification is given" in str(context.exception).lower())
                for station in station_reserved:
                    self.assertTrue(station in str(context.exception).lower())

    def test_create_task_blueprints_and_subtasks_from_scheduling_unit_draft(self):
        """
        Create Task Blueprints and Subtasks (class setup)
        Check if tasks (8) are created:
           Calibration 1     : 1 Observation and 1 Pipeline task
           Target Observation: 1 Observation and 2 Pipeline tasks
           Calibration 2     : 1 Observation and 1 Pipeline task
           Ingest            : 1 ingest task
        Check if subtasks (13) are created:
           Every Observation Task: 3 subtasks (1 control, 2 QA)
           Every Pipeline Task:    1 subtasks (1 control)
           Every Ingest Task:      1 subtasks (1 control)
           makes 3x3 + 4x1 = 13
        """
        self.assertEqual(8, len(self.task_drafts))
        self.assertEqual(1, len(self.scheduling_unit_blueprints))
        self.assertEqual(8, len(self.task_blueprints))
        total_subtasks = 0
        for task_blueprint in self.task_blueprints:
            total_subtasks += task_blueprint.subtasks.count()
        self.assertEqual(14, total_subtasks)

    def test_relative_times(self):
        """
        Create Task Blueprints and Subtasks (class setup)
        Set start and stop times of taskBlueprint
        Set the subtask start/stop time equal to its taskBlueprint
        Set all subtask states to 'finished'
        Check the observed_end_time of the SchedulingUnitBlueprint
        Check the relative_start/stop_time of the SchedulingUnitBlueprint
           start = 0
           stop = calculates like 8hours (Target) + 2x10min (calibrators) + 2*1min (offset between observations) = 8h22min
        """
        DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
        test_timeschedule = {
            # name of taskBlueprint       start_time             stop_time
            "Calibrator Observation 1": ["2020-11-01 08:00:00", "2020-11-01 08:10:00"],
            "Pipeline 1":               ["2020-11-01 08:20:00", "2020-11-01 08:22:00"],
            "Target Observation":       ["2020-11-01 08:30:00", "2020-11-01 18:00:00"],
            "Pipeline target1":         ["2020-11-01 18:30:00", "2020-11-01 18:35:00"],
            "Pipeline target2":         ["2020-11-01 18:40:00", "2020-11-01 18:45:00"],
            "Calibrator Observation 2": ["2020-11-01 19:00:00", "2020-11-01 19:20:00"],
            "Pipeline 2":               ["2020-11-01 19:30:00", "2020-11-01 19:40:00"]
        }
        # Set time_schedule,
        for name, times in test_timeschedule.items():
            task_blueprint = list(filter(lambda x: x.name == name, self.task_blueprints))[0]
            for subtask in task_blueprint.subtasks.all():
                subtask.state = models.SubtaskState.objects.get(value="finished")
                subtask.stop_time = datetime.strptime(times[1], DATETIME_FORMAT)
                subtask.start_time = datetime.strptime(times[0], DATETIME_FORMAT)
                subtask.save()

        # Check times
        self.assertEqual("2020-11-01 19:20:00", self.scheduling_unit_blueprint.observed_end_time.strftime("%Y-%m-%d %H:%M:%S"))
        self.assertEqual(timedelta(0), self.scheduling_unit_blueprint.relative_start_time)
        self.assertEqual(timedelta(hours=8, minutes=22), self.scheduling_unit_blueprint.relative_stop_time)

        for task_blueprint in self.task_blueprints:
            if task_blueprint.name == "Calibrator Observation 1":
                self.assertEqual(timedelta(0), task_blueprint.relative_start_time)
                self.assertEqual(timedelta(minutes=10), task_blueprint.relative_stop_time)
            elif task_blueprint.name == "Target Observation":
                self.assertEqual(timedelta(minutes=11), task_blueprint.relative_start_time)
                self.assertEqual(timedelta(hours=8, minutes=11), task_blueprint.relative_stop_time)
            elif task_blueprint.name == "Calibrator Observation 2":
                self.assertEqual(timedelta(hours=8, minutes=12), task_blueprint.relative_start_time)
                self.assertEqual(timedelta(hours=8, minutes=22), task_blueprint.relative_stop_time)
            else:
                self.assertEqual(timedelta(0), task_blueprint.relative_start_time)
                self.assertEqual(timedelta(0), task_blueprint.relative_stop_time)

    def test_dutch_stations_conflicts_exception(self):
        """
        Test conflict of 'Dutch' station which are have a default of max_nr_missing=4,
        Assign stations equal to max_nr_missing+1 before schedule it and check if it can NOT be scheduled
        Check the context of the Exception
        """
        station_reserved = ['CS002', 'CS003', 'CS004', 'CS401', 'CS501']
        self.assertTrue(create_reserved_stations_for_testing(station_reserved))
        self._schedule_subtask_with_failure(station_reserved)

    def test_dutch_stations_conflicts_ok(self):
        """
        Test conflict of 'Dutch' station which are have a default of max_nr_missing=4,
        Assign stations equal to max_nr_missing before schedule it and check if it can be scheduled
        """
        station_reserved = ['CS002', 'CS003', 'CS004', 'CS401']
        self.assertTrue(create_reserved_stations_for_testing(station_reserved))
        with tmss_test_env.create_tmss_client() as client:
            client.schedule_subtask(self.subtask_id_of_first_observation)

    def test_international_stations_conflicts_failed(self):
        """
        Test conflict of 'International' stations which are have a default of max_nr_missing=2,
        Assign stations equal to max_nr_missing+1 before schedule it and check if it can NOT be scheduled
        Check the context of the Exception
        """
        station_reserved = ['SE607', 'PL610', 'PL612']
        self.assertTrue(create_reserved_stations_for_testing(station_reserved))
        self._schedule_subtask_with_failure(station_reserved)

    def test_international_stations_conflicts_ok(self):
        """
        Test conflict of 'International' stations which are have a default of max_nr_missing=2,
        Assign stations equal to max_nr_missing before schedule it and check if it can be scheduled
        """
        station_reserved = ['SE607', 'PL612']
        self.assertTrue(create_reserved_stations_for_testing(station_reserved))
        with tmss_test_env.create_tmss_client() as client:
            client.schedule_subtask(self.subtask_id_of_first_observation)

    def test_international_required_stations_conflicts_failed(self):
        """
        Test conflict of 'International Required' stations which are have a default of max_nr_missing=1,
        Assign stations equal to max_nr_missing+1 before schedule it and check if it can NOT be scheduled
        Check the context of the Exception
        """
        station_reserved = ['DE601', 'DE605']
        self.assertTrue(create_reserved_stations_for_testing(station_reserved))
        self._schedule_subtask_with_failure(station_reserved)

    def test_international_required_stations_conflicts_ok(self):
        """
        Test conflict of 'International Required' stations which are have a default of max_nr_missing=1,
        Assign stations equal to max_nr_missing before schedule it and check if it can be scheduled
        """
        station_reserved = ['DE605']
        self.assertTrue(create_reserved_stations_for_testing(station_reserved))
        with tmss_test_env.create_tmss_client() as client:
            client.schedule_subtask(self.subtask_id_of_first_observation)


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
