#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas()

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.common.json_utils import resolved_refs, validate_json_against_schema, get_default_json_object_for_schema

class TestSchemas(unittest.TestCase):
    def check_schema(self, name: str, schema: dict):
        """ Check whether the given schema is valid. """

        # Can all $refs be actually resolved?
        logger.info("Resolving references for schema %s", name)
        resolved_refs(schema)

        # Does this schema provide actually valid defaults?
        logger.info("Validating defaults of schema %s", name)
        defaults = get_default_json_object_for_schema(schema)
        validate_json_against_schema(defaults, schema)

    def check_schema_table(self, model):
        """ Check all schemas present in the database for a given model. """

        schemas = model.objects.all()

        for schema in schemas:
            self.check_schema(schema.name, schema.schema)

    def test_subtasks(self):
        self.check_schema_table(models.SubtaskTemplate)

    def test_dataproducts(self):
        self.check_schema_table(models.DataproductSpecificationsTemplate)
        self.check_schema_table(models.DataproductFeedbackTemplate)
        self.check_schema_table(models.SAPTemplate)

    def test_tasks(self):
        self.check_schema_table(models.TaskTemplate)
        self.check_schema_table(models.TaskRelationSelectionTemplate)

    def test_scheduling_units(self):
        self.check_schema_table(models.SchedulingUnitTemplate)
        self.check_schema_table(models.SchedulingConstraintsTemplate)

    def test_reservations(self):
        self.check_schema_table(models.ReservationTemplate)

if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
