#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
import requests

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
rest_data_creator = TMSSRESTTestDataCreator(tmss_test_env.django_server.url, (tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password))

from lofar.sas.tmss.tmss.tmssapp.tasks import *


class CreationFromSchedulingUnitDraft(unittest.TestCase):
    """
    From scheduling_unit_draft should test:
    1. create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    6. create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> [TaskDraft]:
    3. create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    """
    def test_create_scheduling_unit_blueprint_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if NO tasks are created
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")
        strategy_template.template['tasks'] = {}

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   requirements_doc=strategy_template.template,
                                   requirements_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   copy_reason=models.CopyReason.objects.get(value='template'),
                                   generator_instance_doc="para",
                                   copies=None,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
        self.assertEqual(scheduling_unit_draft.name, scheduling_unit_blueprint.draft.name)
        self.assertEqual(0, len(scheduling_unit_blueprint.task_blueprints.all()))

    def test_create_task_drafts_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft (with empty requirements_doc)
        Check if NO tasks are created
        Check with REST-call if NO tasks are created
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")
        strategy_template.template['tasks'] = {}

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   requirements_doc=strategy_template.template,
                                   requirements_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   copy_reason=models.CopyReason.objects.get(value='template'),
                                   generator_instance_doc="para",
                                   copies=None,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        with self.assertRaises(BlueprintCreationException):
            create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft.refresh_from_db()
        task_drafts = scheduling_unit_draft.task_drafts.all()
        self.assertEqual(0, len(task_drafts))
        self.assertEqual(0, len(task_drafts))

    def test_create_task_drafts_from_scheduling_unit_draft_with_UC1_requirements(self):
        """
        Create Scheduling Unit Draft with requirements_doc (read from file)
        Create Task Blueprints (only)
        Check if tasks (7) are created
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   requirements_doc=strategy_template.template,
                                   requirements_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   copy_reason=models.CopyReason.objects.get(value='template'),
                                   generator_instance_doc="para",
                                   copies=None,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft.refresh_from_db()
        task_drafts = scheduling_unit_draft.task_drafts.all()
        self.assertEqual(8, len(task_drafts))

    def test_create_task_blueprints_and_subtasks_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft with empty task specification
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if NO tasks are created
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.get(name="UC1 CTC+pipelines")
        strategy_template.template['tasks'] = {}

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   requirements_doc=strategy_template.template,
                                   requirements_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   copy_reason=models.CopyReason.objects.get(value='template'),
                                   generator_instance_doc="para",
                                   copies=None,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        with self.assertRaises(BlueprintCreationException):
            create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        self.assertEqual(0, len(scheduling_unit_draft.scheduling_unit_blueprints.all()))


class CreationFromSchedulingUnitBluePrint(unittest.TestCase):
    """
    From scheduling_unit_blueprint should test:
    5. create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    """

    def test_create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(self):
        """
        Create Scheduling Unit BluePrint
        Check with REST-call if NO tasks are created, an Exception is raised becaus the requirements_doc of the
        scheduling_unit (draft) has no tasks defined, it is an empty list
        """
        scheduling_unit_blueprint_data = SchedulingUnitBlueprint_test_data(name="Test Scheduling Unit BluePrint")
        scheduling_unit_blueprint = models.SchedulingUnitBlueprint.objects.create(**scheduling_unit_blueprint_data)

        with self.assertRaises(BlueprintCreationException):
            scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)

        self.assertEqual(0, scheduling_unit_blueprint.task_blueprints.count())


class CreationFromTaskDraft(unittest.TestCase):
    """
    From task draft should test:
     2. create_task_blueprint_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
     5. create_task_blueprint_and_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    """
    @staticmethod
    def create_task_object(task_draft_name, output_pinned):
        """
        Helper function to create a task object for testing
        """
        obs_task_template = models.TaskTemplate.objects.get(name='target observation')
        task_draft_data = TaskDraft_test_data(name=task_draft_name, specifications_template=obs_task_template, output_pinned=output_pinned)
        models.TaskDraft.objects.create(**task_draft_data)

    def test_create_task_blueprint_and_subtasks(self):
        """
        Create task draft
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if 3 subtasks are created and if these subtasks have state value 'defined'
        Check that the task blueprint gets created with same output_pinned flag as the task draft.
        """
        self.create_task_object(task_draft_name="Test Target Observation 1", output_pinned=False)

        task_draft = models.TaskDraft.objects.get(name="Test Target Observation 1")
        task_blueprint = create_task_blueprint_and_subtasks_from_task_draft(task_draft)
        self.assertEqual(task_draft.name, task_blueprint.draft.name)
        self.assertEqual(3, task_blueprint.subtasks.count())
        for subtask in task_blueprint.subtasks.all():
            subtask.state.value == 'defined'
        self.assertEqual(task_draft.output_pinned, task_blueprint.output_pinned)

    def test_create_task_blueprint(self):
        """
        Create task draft
        Check if the task draft name is equal to the task draft name specified in the created task blueprint
        Check with REST-call if NO subtasks are created
        Check that the task blueprint gets created with same output_pinned flag as the task draft.
        """
        self.create_task_object(task_draft_name="Test Target Observation 2", output_pinned=True)

        task_draft = models.TaskDraft.objects.get(name="Test Target Observation 2")
        task_blueprint = create_task_blueprint_from_task_draft(task_draft)
        self.assertEqual(task_draft.name, task_blueprint.draft.name)
        self.assertEqual(0, task_blueprint.subtasks.count())
        self.assertEqual(task_draft.output_pinned, task_blueprint.output_pinned)


class TaskBlueprintStateTest(unittest.TestCase):
    """
    Test the Task Blueprint State which is derived from the SubTask states.
    The result of each possible combination of these states will be checked
    See https://support.astron.nl/confluence/display/TMSS/Specification+Flow#SpecificationFlow-TaskBlueprints
    """

    def test_state_with_no_subtasks(self):
        """
        Test the taskblueprint state when subtasks are not instantiated.
        the expected state should be 'defined'
        """
        task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint No Subtasks")
        task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)
        self.assertEqual("defined", task_blueprint.status)

    def test_states_with_one_subtask(self):
        """
        Test the taskblueprint state when only one subtasks is instantiated, an pipeline
        See next table where every row represents:
            Substate(Pipeline), Expected TaskBlueprint State
        """
        test_table = [
            ("defining",    "defined"),
            ("defining",    "defined"),
            ("defined",     "schedulable"),
            ("scheduling",  "schedulable"),
            ("scheduled",   "scheduled"),
            ("starting",    "started"),
            ("started",     "started"),
            ("queueing",    "started"),
            ("queued",      "started"),
            ("finishing",   "started"),
            ("finished",    "finished"),
            ("cancelling",  "cancelled"),
            ("cancelled",   "cancelled"),
            ("error",       "error")
        ]
        # Create taskblueprint
        task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint With One Subtask")
        task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)
        # Create pipeline subtask related to taskblueprint
        subtask_data = Subtask_test_data(task_blueprint, state=models.SubtaskState.objects.get(value="defined"),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='pipeline control'))
        subtask_pipe = models.Subtask.objects.create(**subtask_data)

        # Do the actual test
        for test_item in test_table:
            state_pipe, expected_task_state = test_item
            logger.info("Expected test result of substate pipeline='%s' should be '%s'" % (state_pipe, expected_task_state))
            subtask_pipe.state = models.SubtaskState.objects.get(value=state_pipe)
            subtask_pipe.save()
            self.assertEqual(expected_task_state, task_blueprint.status)

    def test_states_with_observation_and_qa_subtask(self):
        """
        Test the taskblueprint state when two subtasks are instantiated, an observation and a QA.
        See next table where every row represents:
            Substate(Obs), Substate(QA), Expected TaskBlueprint State
        """
        test_table = [
            ("defining",    "defining",   "defined"),
            ("defining",    "defined",    "defined"),
            ("defined",     "defined",    "schedulable"),
            ("scheduling",  "defined",    "schedulable"),
            ("scheduled",   "defined",    "scheduled"),
            ("starting",    "defined",    "started"),
            ("started",     "defined",    "started"),
            ("queueing",    "defined",    "started"),
            ("queued",      "defined",    "started"),
            ("finishing",   "defined",    "observed"),
            ("finished",    "defined",    "observed"),
            ("finished",    "finished",   "finished"),
            ("cancelling",  "defined",    "cancelled"),
            ("cancelled",   "defined",    "cancelled"),
            ("error",       "defined",    "error"),
            # qa finishing/finished should be not observed
            ("defined",     "finishing",  "started"),
            ("defined",     "finished",   "started"),
            ("scheduled",   "finishing",  "started"),
            ("scheduled",   "finished",   "started"),
            # error and cancelled/ing
            ("scheduled",   "error",      "error"),
            ("scheduled",   "cancelling", "cancelled"),
            ("scheduled",   "cancelled",  "cancelled"),
            ("started",     "error",      "error"),
            ("started",     "cancelling", "cancelled"),
            ("started",     "cancelled",  "cancelled"),
            ("finished",    "error",      "error"),
            ("finished",    "cancelling", "cancelled"),
            ("finished",    "cancelled",  "cancelled"),
            # cancelled over error
            ("cancelling",  "error",      "cancelled"),
            ("cancelled",   "error",      "cancelled"),
            ("error",       "cancelling", "cancelled"),
            ("error",       "cancelling", "cancelled"),
            # qa scheduled
            ("starting",    "scheduled",  "started"),
            ("started",     "scheduled",  "started"),
            ("queueing",    "scheduled",  "started"),
            ("queued",      "scheduled",  "started"),
            ("finishing",   "scheduled",  "observed"),
            ("finished",    "scheduled",  "observed"),
            ("cancelling",  "scheduled", "cancelled"),
            ("cancelled",   "scheduled", "cancelled"),
            ("error",       "scheduled", "error"),
        ]
        # Create taskblueprint
        task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint With Subtasks")
        task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)
        # Create observation and qa subtask related to taskblueprint
        subtask_data = Subtask_test_data(task_blueprint, state=models.SubtaskState.objects.get(value="defined"),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='observation control'))
        subtask_obs = models.Subtask.objects.create(**subtask_data)
        subtask_data = Subtask_test_data(task_blueprint, state=models.SubtaskState.objects.get(value="defined"),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='QA file conversion'))
        subtask_qa = models.Subtask.objects.create(**subtask_data)

        # Do the actual test
        for test_item in test_table:
            state_obs, state_qa, expected_task_state = test_item
            logger.info("Expected test result of substates observation='%s' and qa='%s' should be '%s'" % (state_obs, state_qa, expected_task_state))
            subtask_obs.state = models.SubtaskState.objects.get(value=state_obs)
            subtask_obs.save()
            subtask_qa.state = models.SubtaskState.objects.get(value=state_qa)
            subtask_qa.save()
            self.assertEqual(expected_task_state, task_blueprint.status)

    def test_states_with_two_observation_and_two_qa_subtasks(self):
        """
        Test the taskblueprint state when four subtasks are instantiated, two observation and two QA.
        See next table where every row represents:
            Substate(Obs1), Substate(Obs2), Substate(QA1), Substate(QA2), Expected TaskBlueprint State
        """
        test_table = [
            ("finishing",   "defined",    "defined",    "defined",    "started"),
            ("finished",    "defined",    "defined",    "defined",    "started"),
            ("finishing",   "started",    "defined",    "defined",    "started"),
            ("finished",    "started",    "defined",    "defined",    "started"),
            ("finishing",   "finishing",  "defined",    "defined",    "observed"),
            ("finished",    "finished",   "defined",    "defined",    "observed"),
            ("finished",    "finished",   "scheduled",  "defined",    "observed"),
            ("finished",    "finished",   "finished",   "scheduled",  "observed"),
            ("finished",    "finished",   "finished",   "finished",   "finished"),
            ("finished",    "finished",   "finished",   "cancelled",  "cancelled"),
            ("finished",    "finished",   "finished",   "error",      "error"),
            ("error",       "finished",   "finished",   "cancelled",  "cancelled"),
        ]
        # Create taskblueprint
        task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint With Subtasks")
        task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)
        # Create observation and qa subtasks related to taskblueprint
        subtask_data = Subtask_test_data(task_blueprint, state=models.SubtaskState.objects.get(value="defined"),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='observation control'))
        subtask_obs1 = models.Subtask.objects.create(**subtask_data)
        subtask_obs2 = models.Subtask.objects.create(**subtask_data)
        subtask_data = Subtask_test_data(task_blueprint, state=models.SubtaskState.objects.get(value="defined"),
                                         subtask_template=models.SubtaskTemplate.objects.get(name='QA file conversion'))
        subtask_qa1 = models.Subtask.objects.create(**subtask_data)
        subtask_qa2 = models.Subtask.objects.create(**subtask_data)

        # Do the actual test
        for test_item in test_table:
            state_obs1, state_obs2, state_qa1, state_qa2, expected_task_state = test_item
            logger.info("Expected test result of substates observation='%s','%s' and qa='%s','%s' should be '%s'" %
                        (state_obs1, state_obs1, state_qa1, state_qa2, expected_task_state))
            subtask_obs1.state = models.SubtaskState.objects.get(value=state_obs1)
            subtask_obs1.save()
            subtask_obs2.state = models.SubtaskState.objects.get(value=state_obs2)
            subtask_obs2.save()
            subtask_qa1.state = models.SubtaskState.objects.get(value=state_qa1)
            subtask_qa1.save()
            subtask_qa2.state = models.SubtaskState.objects.get(value=state_qa2)
            subtask_qa2.save()
            self.assertEqual(expected_task_state, task_blueprint.status)



if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
