#!/bin/bash
set +x

# get free port for postgres (default port or first subsequent free)
function get_free_port {
  comm -23 <(seq $1 65535) <(ss -tan | awk '{print $4}' | cut -d':' -f2 | grep "[0-9]\{1,5\}" | sort | uniq) | head -n 1
}

function setup {
    # write test credentials to file
    mkdir -p ~/.lofar/dbcredentials

    DJANGO_TEST_DATABASE_PORT=$(get_free_port 7531)
    DJANGO_TEST_DATABASE_NAME="tmss_test_db_`uuidgen | sed 's/-/_/g'`"
    DJANGO_TEST_DATABASE_CREDENTIALS_PATH="$HOME/.lofar/dbcredentials/$DJANGO_TEST_DATABASE_NAME.ini"

    echo "creating test db credentials file: $DJANGO_TEST_DATABASE_CREDENTIALS_PATH"
    echo "[database:$DJANGO_TEST_DATABASE_NAME]
    host=localhost
    type=postgres
    database=$DJANGO_TEST_DATABASE_NAME
    port=$DJANGO_TEST_DATABASE_PORT
    user=tmss
    password=tmss
    " > "$DJANGO_TEST_DATABASE_CREDENTIALS_PATH"


    DJANGO_TEST_LDAP_PORT=$(get_free_port 8642)
    DJANGO_TEST_LDAP_NAME="tmss_test_ldap_`uuidgen | sed 's/-/_/g'`"
    DJANGO_TEST_LDAP_CREDENTIALS_PATH="$HOME/.lofar/dbcredentials/$DJANGO_TEST_LDAP_NAME.ini"

    echo "creating test ldap credentials file: $DJANGO_TEST_LDAP_CREDENTIALS_PATH"
    echo "[database:$DJANGO_TEST_LDAP_NAME]
    port=$DJANGO_TEST_LDAP_PORT
    user=tmss_test_ldap_user
    password=tmss_test_ldap_password
    " > "$DJANGO_TEST_LDAP_CREDENTIALS_PATH"


    #keep track of helper application pids to kill in teardown
    PIDS=

    DJANGO_TEST_PORT=$(get_free_port 8765)

    export DJANGO_TEST_PORT=$DJANGO_TEST_PORT
    export TMSS_DBCREDENTIALS=$DJANGO_TEST_DATABASE_NAME
    export TMSS_LDAPCREDENTIALS=$DJANGO_TEST_LDAP_NAME

    ## Trap upon signals and upon normal exit.
    trap 'STATUS=$?; teardown; exit $STATUS' SIGHUP SIGINT SIGQUIT SIGKILL SIGTERM
}

function teardown {
    echo "tearing down test environment"
    for PID in $PIDS
    do
        # get the full command
        CMD="`ps --pid $PID h -o command`"
        echo "killing helper application: pid=$PID cmd='$CMD'"

        # and kill it
        kill -TERM $PID
    done

    echo "removing test db credentials file: $DJANGO_TEST_DATABASE_CREDENTIALS_PATH"
    rm $DJANGO_TEST_DATABASE_CREDENTIALS_PATH

    echo "removing test ldap credentials file: $DJANGO_TEST_LDAP_CREDENTIALS_PATH"
    rm $DJANGO_TEST_LDAP_CREDENTIALS_PATH
}


function setup_with_test_tmss {
    setup

    # fire up a postgres test database
    echo
    echo "Starting Django test database server..."
    tmss_testdatabase -C $DJANGO_TEST_DATABASE_NAME &
    PIDS="$! $PIDS"
    sleep 5
    echo "Started Django test database server"
    echo

    # fire up a test LDAP service
    echo
    echo "Starting test LDAP server..."
    tmss_testldap -C $TMSS_LDAPCREDENTIALS &
    PIDS="$! $PIDS"
    sleep 5
    echo "Started test LDAP server"
    echo

    # Run Django test instance
    echo
    echo "Starting tmss django server..."
    export TMSS_RAISE_ON_SIGNALS="True"
    tmss -p $DJANGO_TEST_PORT -C $DJANGO_TEST_DATABASE_NAME &
    PIDS="$! $PIDS"
    sleep 5
    echo "Started tmss django server"
    echo
}

function run_test {
    # Run test
    echo "starting test: $1"
    $1 &
    TEST_PID=$!

    wait "$TEST_PID"
    # wait again (to get the status code of the test)
    wait "$TEST_PID"
    TEST_EXIT_CODE=$?

    teardown

    exit $TEST_EXIT_CODE
}

