import logging
logger = logging.getLogger(__name__)

import json
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession
from lofar.common import json_utils
import os
from concurrent.futures import ThreadPoolExecutor

def populate_schemas_main():
    from optparse import OptionParser

    # Check the invocation arguments
    parser = OptionParser('%prog [options]', description='upload the templates to TMSS')
    parser.add_option('-d', '--dir', dest='schema_dir', type='string',
                      default=os.path.expandvars('$LOFARROOT/share/tmss/schemas'),
                      help='''directory path containing the schemas, default: '%default')''')
    parser.add_option('-f', '--file', dest='templates_file', type='string',
                      default='templates.json',
                      help='''json file containing the list of templates with name, description, vesions, and template type. default: '%default')''')
    (options, args) = parser.parse_args()
    return populate_schemas(options.schema_dir, options.templates_file)

def populate_schemas(schema_dir: str=None, templates_filename: str=None):
    with TMSSsession.create_from_dbcreds_for_ldap() as client:
        if schema_dir is None:
            schema_dir = os.path.expandvars('$LOFARROOT/share/tmss/schemas')

        if templates_filename is None:
            templates_filename = 'templates.json'

        templates_filepath = os.path.join(schema_dir, templates_filename)
        logger.info("Reading templates in: %s", templates_filepath)

        with open(templates_filepath) as templates_file:
            templates = json.loads(templates_file.read())

            # keep track of the templates, json schemas and references
            templates_dict = {}
            observing_strategy_templates = []
            schema_references = {}
            all_references = set()

            # load all templates and schemas and prepare them for upload.
            # determine the dependencies, and upload the depenends first, and the rest in parallel later.
            for template in templates:
                with open(os.path.join(schema_dir, template['file_name'])) as schema_file:
                    json_schema = json.loads(schema_file.read())

                # add template name/description/version from schema if not already in template
                template['name'] = template.get('name', json_schema.get('title', '<no name>'))
                template['description'] = template.get('description', json_schema.get('description', '<no description>'))
                template['version'] = template.get('version', '1')

                template_name = template['template']

                if template_name == 'subtask_template' and 'type' in template:
                    # override plain-text type by its url
                    template['type'] = client.get_full_url_for_path('subtask_type/' + template.get('type'))

                if template_name == 'task_template' and 'type' in template:
                    # override plain-text type by its url
                    template['type'] = client.get_full_url_for_path('task_type/' + template.get('type'))

                # inject a unique id in the form of a unique URL to this schema
                json_schema['$id'] = client.get_full_url_for_path('schemas/%s/%s/%s' % (template_name.replace('_', ''), template['name'], template['version']))

                # make sure that all urls point to the tmss base_url
                json_schema = json_utils.replace_host_in_urls(json_schema, new_base_url=client.host_url)

                # get the id without trailing # and/or /
                json_schema_id = json_schema.get('$id', "").rstrip("#").rstrip("/")

                if template_name == 'scheduling_unit_observing_strategy_template':
                    template['template'] = json_schema
                else:
                    template['schema'] = json_schema

                # what are the references? on which other schema's does this schema depend?
                refs = set(ref[:ref.find('#')].rstrip('/') for ref in json_utils.get_refs(json_schema) if not ref.startswith(json_schema_id) and ref.startswith("http"))
                schema_references[json_schema_id] = refs
                all_references.update(refs)

                # store the prepared template for upload
                if template_name == 'scheduling_unit_observing_strategy_template':
                    observing_strategy_templates.append(template)
                else:
                    templates_dict[json_schema_id] = template

            # helper functions for uploading
            def upload_template(template: dict):
                logger.info("Uploading template with name='%s' version='%s'", template['name'], template['version'])
                client.post_template(template_path=template.pop('template'), **template)

            # helper functions for uploading
            def upload_template_if_needed_with_dependents_first(id: str):
                if id in templates_dict:
                    #recurse over dependents if any
                    refs = schema_references.get(id, [])
                    for ref in refs:
                        upload_template_if_needed_with_dependents_first(ref)

                    template = templates_dict.pop(id)
                    upload_template(template)

            # helper functions for uploading observing_strategy_templates
            def upload_observing_strategy_templates(template: dict):
                scheduling_unit_templates = client.get_path_as_json_object('scheduling_unit_template?name=' + template.get('scheduling_unit_template_name') + '&version=' + template.get('scheduling_unit_template_version'))
                scheduling_unit_template = scheduling_unit_templates[0]
                template['scheduling_unit_template'] = scheduling_unit_template['url']
                logger.info("Uploading observation strategy with name='%s' version='%s'", template['name'], template['version'])
                client.post_template(template_path='scheduling_unit_observing_strategy_template', **template)

            # first, upload all dependent templates
            for ref in all_references:
                upload_template_if_needed_with_dependents_first(ref)

            # then, upload the remaining templates in parallel
            rest_templates = [template for template in templates_dict.values()]
            with ThreadPoolExecutor() as executor:
               executor.map(upload_template, rest_templates)

            # and finally, the observing_strategy_templates
            with ThreadPoolExecutor() as executor:
               executor.map(upload_observing_strategy_templates, observing_strategy_templates)

            scheduling_constraints_templates = client.get_path_as_json_object('scheduling_constraints_template')
            if scheduling_constraints_templates:
                default_scheduling_constraints_template = scheduling_constraints_templates[0]
                logger.info("Making scheduling_constraints_templates name='%s' version='%s' the default", default_scheduling_constraints_template['name'], default_scheduling_constraints_template['version'])
                client.session.post(client.get_full_url_for_path('default_scheduling_constraints_template'), json={'name': default_scheduling_constraints_template['name'],
                                                                                                                   'template': default_scheduling_constraints_template['url']})

